/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 *  Authors: Jeffrey Stedfast <fejj@helixcode.com>
 *
 *  Copyright 2000 Helix Code, Inc. (www.helixcode.com)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Street #330, Boston, MA 02111-1307, USA.
 *
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <gnome.h>
#include <glade/glade.h>
#include <gmime/gmime.h>
#include <libgnomevfs/gnome-vfs.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "spruce-composer.h"
#include "spruce-composer-attachment.h"


#define DEFAULT_WIDTH  500
#define DEFAULT_HEIGHT 550


enum {
	SEND,
	POSTPONE,
	LAST_SIGNAL
};

enum {
	ADDRESSBOOK_TO,
	ADDRESSBOOK_CC,
	ADDRESSBOOK_BCC
};

typedef enum {
	MSG_FORMAT_PLAIN,
	MSG_FORMAT_ALTERNATIVE,
} MsgFormat;

static guint signals[LAST_SIGNAL] = { 0 };

static GnomeAppClass *parent_class = NULL;


static gchar *
get_mime_type (const gchar *file_name)
{
	GnomeVFSFileInfo info;
	GnomeVFSResult result;

	result = gnome_vfs_get_file_info (file_name, &info,
					  GNOME_VFS_FILE_INFO_GET_MIME_TYPE |
					  GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
	if (result == GNOME_VFS_OK) {
		gchar *type;

		type = g_strdup (gnome_vfs_file_info_get_mime_type (&info));
		gnome_vfs_file_info_unref (&info);
		return type;
	} else
		return NULL;
}

static GMimeMessage *
build_message (SpruceComposer *composer)
{
	MsgFormat type = MSG_FORMAT_ALTERNATIVE;
	GMimeMessage *message;
	GMimePart *body;
	GMimePart *part;
	gchar *from = NULL;
	gboolean e8bit;
	char *plain = NULL;
	int i;
	
	if (composer->send_html)
		type = MSG_FORMAT_ALTERNATIVE;
	else
		type = MSG_FORMAT_PLAIN;
	
	message = g_mime_message_new ();
	
	/* FIXME: fill in all the parts */
	
	return message;
}

static char *
get_signature (const char *sigfile)
{
	GString *rawsig;
	char buf[1024];
	char *sig = NULL;
	int fd, n;
	
	if (!sigfile || !*sigfile) {
		return NULL;
	}
	
	fd = open (sigfile, O_RDONLY);
	if (fd == -1) {
		char *msg;
		
		msg = g_strdup_printf ("Could not open signature file %s:\n%s",
				       sigfile, g_strerror (errno));
		gnome_error_dialog (msg);
		g_free (msg);
		
		return NULL;
	}
	
	rawsig = g_string_new ("");
	while ((n = read (fd, buf, 1023)) > 0) {
		buf[n] = '\0';
		g_string_append (rawsig, buf);
	}
	close (fd);
	
	sig = rawsig->str;
	g_string_free (rawsig, TRUE);
	
	return sig;
}

static void
set_editor_text (GtkWidget *editor, const char *sig_file, const char *text)
{
	char *sig, *fulltext;
	guint len;
	
	sig = get_signature (sig_file);
	if (sig) {
		if (!strncmp ("-- \n", sig, 3))
			fulltext = g_strdup_printf ("%s\n%s", text, sig);
		else
			fulltext = g_strdup_printf ("%s\n-- \n%s", text, sig);
	} else {
		if (!*text)
			return;
		fulltext = g_strdup (text);
	}
	
	g_free (sig);
	
	gtk_text_freeze (GTK_TEXT (editor));
	len = gtk_text_get_length (GTK_TEXT (editor));
	gtk_text_set_point (GTK_TEXT (editor), len);
	gtk_text_backward_delete (GTK_TEXT (editor), len);
	/* FIXME: use user's prefered font */
	gtk_text_insert (GTK_TEXT (editor), NULL, NULL, NULL, fulltext, -1);
	gtk_text_thaw (GTK_TEXT (editor));
	
	g_free (fulltext);
}

static void
exit_dialog_cb (int reply, SpruceComposer *composer)
{
	GMimeMessage *message;
	
	switch (reply) {
	case 0:
		/* FIXME: save the mime message in Drafts */
		/*message = build_message (composer);
		  mailbox_append_message (DRAFTS_FOLDER, message);
		  g_mime_message_destroy (message);*/
	case 1:
		gtk_widget_destroy (GTK_WIDGET (composer));
		break;
	case 2:
	default:
	}
}

static void
do_exit (SpruceComposer *composer)
{
	GtkWidget *dialog;
	GtkWidget *label;
	gint button;
	
	dialog = gnome_dialog_new (_("Spruce"),
				   GNOME_STOCK_BUTTON_YES,      /* Save */
				   GNOME_STOCK_BUTTON_NO,       /* Don't save */
				   GNOME_STOCK_BUTTON_CANCEL,   /* Cancel */
				   NULL);
	
	label = gtk_label_new (_("This message has not been sent.\n\nDo you wish to save your changes?"));
	gtk_box_pack_start (GTK_BOX (GNOME_DIALOG (dialog)->vbox), label, TRUE, TRUE, 0);
	gtk_widget_show (label);
	gnome_dialog_set_parent (GNOME_DIALOG (dialog), GTK_WINDOW (composer));
	gnome_dialog_set_default (GNOME_DIALOG (dialog), 0);
	button = gnome_dialog_run_and_close (GNOME_DIALOG (dialog));
	
	exit_dialog_cb (button, composer);
}

static void
on_mnuFileSend_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	gtk_signal_emit (GTK_OBJECT (user_data), signals[SEND]);
}

static void
on_cmdSend_clicked (GtkButton *button, gpointer user_data)
{
	gtk_signal_emit (GTK_OBJECT (user_data), signals[SEND]);
}


static void
on_mnuFileSendLater_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	gtk_signal_emit (GTK_OBJECT (user_data), signals[POSTPONE]);
}


static void
on_mnuFileSave_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	
}


static void
on_cmdSave_clicked (GtkButton *button, gpointer user_data)
{
	
}


static void
on_mnuFileSaveAs_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	
}


static void
on_mnuFilePrint_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	/* TODO: Add code to print... */
}


static void
on_cmdPrint_clicked (GtkButton *button, gpointer user_data)
{
	/* TODO: Add code to print... */
}


static void
on_mnuFileClose_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	SpruceComposer *composer;
	
	composer = SPRUCE_COMPOSER (user_data);
	do_exit (composer);
}


static void
on_mnuEditCut_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	
}


static void
on_mnuEditCopy_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	
}


static void
on_mnuEditPaste_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	
}


static void
on_mnuEditClear_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	
}


static void
on_mnuInsertFile_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	
}


static void
on_mnuInsertAttachment_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	
}


static void
on_cmdAttach_activate (GtkButton *button, gpointer user_data)
{
	
}


static void
on_mnuInsertSignature_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	SpruceComposer *composer;
	char *sig, *fulltext;
	guint len;
	
	composer = SPRUCE_COMPOSER (user_data);
	
	sig = get_signature (composer->sig_file);
	if (sig) {
		if (!strncmp ("-- \n", sig, 3))
			fulltext = g_strdup_printf ("\n%s", sig);
		else
			fulltext = g_strdup_printf ("\n-- \n%s", sig);
		
		g_free (sig);
		
		gtk_text_freeze (GTK_TEXT (composer->editor));
		len = gtk_text_get_length (GTK_TEXT (composer->editor));
		gtk_text_set_point (GTK_TEXT (composer->editor), len);
		gtk_text_backward_delete (GTK_TEXT (composer->editor), len);
		/* FIXME: use user's prefered font */
		gtk_text_insert (GTK_TEXT (composer->editor), NULL, NULL, NULL, fulltext, -1);
		gtk_text_thaw (GTK_TEXT (composer->editor));
		
		g_free (fulltext);
	}
}


static void
on_mnuInsertSpruceSignature_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	
}


static void
on_mnuFormatHTML_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	SpruceComposer *composer = SPRUCE_COMPOSER (user_data);
	
	composer->send_html = !composer->send_html;
}


static void
on_mnuHelpAbout_activate (GtkMenuItem *menuitem, gpointer user_data)
{
	
}


static void
on_cmdAddressbookLookup_clicked (GtkButton *button, gpointer user_data)
{
	
}


static void
destroy (GtkObject *object)
{
	SpruceComposer *composer;
	
	composer = SPRUCE_COMPOSER (object);
	
	if (composer->extra_hdr_names) {
		int i;
		
		for (i = 0; i < composer->extra_hdr_names->len; i++) {
			g_free (composer->extra_hdr_names->pdata[i]);
			g_free (composer->extra_hdr_values->pdata[i]);
		}
		
		g_ptr_array_free (composer->extra_hdr_names, TRUE);
		g_ptr_array_free (composer->extra_hdr_values, TRUE);
	}
}

static int
delete_event (GtkWidget *widget, GdkEventAny *event)
{
	do_exit (SPRUCE_COMPOSER (widget));
	
	return TRUE;
}

static void
class_init (SpruceComposerClass *klass)
{
	GtkObjectClass *object_class;
	GtkWidgetClass *widget_class;
	
	object_class = GTK_OBJECT_CLASS (klass);
	widget_class = GTK_WIDGET_CLASS (klass);
	
	object_class->destroy = destroy;
	
	widget_class->delete_event = delete_event;
	
	parent_class = gtk_type_class (gnome_app_get_type ());
	
	signals[SEND] =
		gtk_signal_new ("send",
				GTK_RUN_LAST,
				object_class->type,
				GTK_SIGNAL_OFFSET (SpruceComposerClass, send),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE, 0);
	
	signals[POSTPONE] =
		gtk_signal_new ("postpone",
				GTK_RUN_LAST,
				object_class->type,
				GTK_SIGNAL_OFFSET (SpruceComposerClass, postpone),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE, 0);
	
	gtk_object_class_add_signals (object_class, signals, LAST_SIGNAL);
}

static void
init (SpruceComposer *composer)
{
	gui                        = NULL;
	
	composer->extra_hdr_names  = g_ptr_array_new ();
	composer->extra_hdr_values = g_ptr_array_new ();
	
	/* widgets */
	composer->from             = NULL;
	composer->to               = NULL;
	composer->cc               = NULL;
	composer->bcc              = NULL;
	composer->subject          = NULL;
	composer->editor           = NULL;
	composer->attachments      = NULL;
	
	composer->send_html        = FALSE;
}

GtkType
spruce_composer_get_type (void)
{
	static GtkType type = 0;
	
	if (type == 0) {
		static const GtkTypeInfo info = {
			"SpruceComposer",
			sizeof (SpruceComposer),
			sizeof (SpruceComposerClass),
			(GtkClassInitFunc) class_init,
			(GtkObjectInitFunc) init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL,
		};
		
		type = gtk_type_unique (gtk_widget_get_type (), &info);
	}
	
	return type;
}


static void
spruce_composer_construct (SpruceComposer *composer)
{
	GladeXML *gui;
	GtkWidget *widget;
	
	g_return_if_fail (IS_SPRUCE_COMPOSER (composer));
	
	gui = glade_xml_new (GLADEDIR "/spruce-composer.glade", NULL);
	
	if (gui == NULL) {
		g_warning ("Cannot load `spruce-composer.glade'");
		return;
	}
	
	composer->gui = gui;
	
	composer->from = glade_xml_get_widget (gui, "cmboFrom");
	composer->to = glade_xml_get_widget (gui, "txtTo");
	composer->cc = glade_xml_get_widget (gui, "txtCc");
	composer->bcc = glade_xml_get_widget (gui, "txtBcc");
	composer->subject = glade_xml_get_widget (gui, "txtSubject");
	
	composer->editor = glade_xml_get_widget (gui, "editor");
	
	composer->attachments = glade_xml_get_widget (gui, "iconlistAttachments");
	
	/* toolbar buttons */
	widget = glade_xml_get_widget (gui, "cmdSend");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_cmdSend_clicked),
			    composer);
	widget = glade_xml_get_widget (gui, "cmdSave");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_cmdSave_clicked),
			    composer);
	widget = glade_xml_get_widget (gui, "cmdPrint");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_cmdPrint_clicked),
			    composer);
	widget = glade_xml_get_widget (gui, "cmdAttach");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_cmdAttach_clicked),
			    composer);
	
	/* File menu */
	widget = glade_xml_get_widget (gui, "mnuFileSend");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuFileSend_activate),
			    composer);
	widget = glade_xml_get_widget (gui, "mnuFileSendLater");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuFileSendLater_activate),
			    composer);
	widget = glade_xml_get_widget (gui, "mnuFileSave");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuFileSave_activate),
			    composer);
	widget = glade_xml_get_widget (gui, "mnuFileSaveAs");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuFileSaveAs_activate),
			    composer);
	widget = glade_xml_get_widget (gui, "mnuFilePrint");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuFilePrint_activate),
			    composer);
	widget = glade_xml_get_widget (gui, "mnuFileClose");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuFileClose_activate),
			    composer);
	
	/* Edit menu */
	widget = glade_xml_get_widget (gui, "mnuEditCut");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuEditCut_activate),
			    composer);
	widget = glade_xml_get_widget (gui, "mnuEditCopy");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuEditCopy_activate),
			    composer);
	widget = glade_xml_get_widget (gui, "mnuEditPaste");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuEditPaste_activate),
			    composer);
	widget = glade_xml_get_widget (gui, "mnuEditClear");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuEditClear_activate),
			    composer);
	widget = glade_xml_get_widget (gui, "mnuInsertFile");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuInsertFile_activate),
			    composer);
	widget = glade_xml_get_widget (gui, "mnuInsertAttachment");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuInsertAttachment_activate),
			    composer);
	widget = glade_xml_get_widget (gui, "mnuInsertSignature");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuInsertSignature_activate),
			    composer);
	widget = glade_xml_get_widget (gui, "mnuInsertSpruceSignature");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuInsertSpruceSignature_activate),
			    composer);
	
	/* Format menu */
	widget = glade_xml_get_widget (gui, "mnuFormatHTML");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuFormatHTML_activate),
			    composer);
	
	/* Help menu */
	widget = glade_xml_get_widget (gui, "mnuHelpAbout");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_mnuHelpAbout_activate),
			    composer);
	
	/* addressbook lookup buttons */
	widget = glade_xml_get_widget (gui, "cmdToAddr");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_cmdAddressbookLookup_clicked),
			    composer->to);
	widget = glade_xml_get_widget (gui, "cmdCcAddr");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_cmdAddressbookLookup_clicked),
			    composer->cc);
	widget = glade_xml_get_widget (gui, "cmdBccAddr");
	gtk_signal_connect (GTK_OBJECt (widget), "activate",
			    GTK_SIGNAL_FUNC (on_cmdAddressbookLookup_clicked),
			    composer->bcc);
	
	/* composer window signals */
	widget = glade_xml_get_widget (gui, "spruce_composer");
	gtk_signal_connect (GTK_OBJECt (widget), "delete_event",
			    GTK_SIGNAL_FUNC (delete_event),
			    composer);
}

static SpruceComposer *
create_composer (void)
{
	SpruceComposer *new;

	new = gtk_type_new (TYPE_SPRUCE_COMPOSER);
	spruce_composer_construct (new);
	if (!new->editor) {
		g_warning (_("Could not create composer window."));
		gtk_object_unref (GTK_OBJECT (new));
		return NULL;
	}
	
	return new;
}


/**
 * spruce_composer_new:
 *
 * Create a new message composer widget.
 * 
 * Return value: A pointer to the newly created widget
 **/
SpruceComposer *
e_msg_composer_new (void)
{
	SpruceComposer *new;
	
	new = create_composer ();
	if (new) {
		/* Load the signature, if any. */
		set_editor_text (GTK_WIDGET (new->editor), NULL, "");
	}
	
	return new;
}

/**
 * spruce_composer_new_with_sig_file:
 *
 * Create a new message composer widget. Sets the signature file.
 * 
 * Return value: A pointer to the newly created widget
 **/
SpruceComposer *
spruce_composer_new_with_sig_file (const char *sig_file)
{
	SpruceComposer *new;
	
	new = create_composer ();
	if (new) {
		/* Load the signature, if any. */
		set_editor_text (GTK_WIDGET (new->editor), sig_file, "");
		
		spruce_composer_set_sig_file (new, sig_file);
	}
	
	return new;
}

/**
 * spruce_composer_new_with_message:
 *
 * Create a new message composer widget.
 * 
 * Return value: A pointer to the newly created widget
 **/
SpruceComposer *
spruce_composer_new_with_message (GMimeMessage *msg)
{
	SpruceComposer *new;
	GList *To, *Cc, *Bcc;
	const gchar *subject;
	gchar *text;
	
	new = create_composer ();
	if (!new)
		return NULL;
	
	To = g_mime_message_get_recipients (msg, G_MIME_MESSAGE_RECIPIETS_TO);
	Cc = g_mime_message_get_recipients (msg, G_MIME_MESSAGE_RECIPIETS_CC);
	Bcc = g_mime_message_get_recipients (msg, G_MIME_MESSAGE_RECIPIETS_BCC);
	subject = g_mime_message_get_subject (msg);
	
	spruce_composer_set_headers (new, To, Cc, Bcc, subject);
	
	text = get_message_body (msg);
	
	spruce_composer_set_body_text (new, text);
	
	return new;
}

/**
 * spruce_composer_set_headers:
 * @composer: a composer object
 * @to: the values for the "To" header
 * @cc: the values for the "Cc" header
 * @bcc: the values for the "Bcc" header
 * @subject: the value for the "Subject" header
 *
 * Sets the headers in the composer to the given values.
 **/
void 
spruce_composer_set_headers (SpruceComposer *composer, const GList *to,
			     const GList *cc, const GList *bcc,
			     const char *subject)
{
	g_return_if_fail (IS_SPRUCE_COMPOSER (composer));
	
	/*spruce_composer_hdrs_set_to (hdrs, to);
	  spruce_composer_hdrs_set_cc (hdrs, cc);
	  spruce_composer_hdrs_set_bcc (hdrs, bcc);
	  spruce_composer_hdrs_set_subject (hdrs, subject);*/
}


/**
 * spruce_composer_set_body_text:
 * @composer: a composer object
 * @text: the text to initialize the editor with
 *
 * Loads the given text into the editor.
 **/
void
spruce_composer_set_body_text (SpruceComposer *composer, const char *text)
{
	g_return_if_fail (IS_SPRUCE_COMPOSER (composer));
	
	set_editor_text (GTK_WIDGET (composer->editor), 
			 composer->sig_file, text);
}

/**
 * spruce_composer_add_header:
 * @composer: a composer object
 * @name: the header name
 * @value: the header value
 *
 * Adds a header with @name and @value to the message. This header
 * may not be displayed by the composer, but will be included in
 * the message it outputs.
 **/
void
spruce_composer_add_header (SpruceComposer *composer, const char *name,
			    const char *value)
{
	g_return_if_fail (IS_SPRUCE_COMPOSER (composer));
	g_return_if_fail (name != NULL);
	g_return_if_fail (value != NULL);
	
	g_ptr_array_add (composer->extra_hdr_names, g_strdup (name));
	g_ptr_array_add (composer->extra_hdr_values, g_strdup (value));
}


/**
 * spruce_composer_attach:
 * @composer: a composer object
 * @attachment: the GMimePart to attach
 *
 * Attaches @attachment to the message being composed in the composer.
 **/
void
spruce_composer_attach (SpruceComposer *composer, GMimePart *attachment)
{
	g_return_if_fail (IS_SPRUCE_COMPOSER (composer));
	g_return_if_fail (attachment != NULL);
	
	/* FIXME: attach the mime part? */
}


/**
 * spruce_composer_get_message:
 * @composer: A message composer widget
 * 
 * Retrieve the message edited by the user as a CamelMimeMessage.  The
 * GMimeMessage object is created on the fly; subsequent calls to this
 * function will always create new objects from scratch.
 * 
 * Return value: A pointer to the new GMimeMessage object
 **/
GMimeMessage *
spruce_composer_get_message (SpruceComposer *composer)
{
	g_return_val_if_fail (IS_SPRUCE_COMPOSER (composer), NULL);
	
	return build_message (composer);
}


/**
 * spruce_composer_set_sig:
 * @composer: A message composer widget
 * @path: Signature file
 * 
 * Set a signature
 **/
void
spruce_composer_set_sig_file (SpruceComposer *composer, const char *sig_file)
{
	g_return_if_fail (IS_SPRUCE_COMPOSER (composer));
	
	composer->sig_file = g_strdup (sig_file);
}

/**
 * spruce_composer_get_sig_file:
 * @composer: A message composer widget
 * 
 * Get the signature file
 * 
 * Return value: The signature file.
 **/
const char *
spruce_composer_get_sig_file (SpruceComposer *composer)
{
	g_return_val_if_fail (IS_SPRUCE_COMPOSER (composer), NULL);
	
	return composer->sig_file;
}

/**
 * spruce_composer_set_send_html:
 * @composer: A message composer widget
 * @send_html: Whether the composer should have the "Send HTML" flag set
 * 
 * Set the status of the "Send HTML" toggle item.  The user can override it.
 **/
void
spruce_composer_set_send_html (SpruceComposer *composer,
			       gboolean send_html)
{
	GtkWidget *item;
	
	g_return_if_fail (IS_SPRUCE_COMPOSER (composer));
	
	composer->send_html = send_html;
	
	item = glade_xml_get_widget (composer->gui, "mnuFormatHTML");
	
	gtk_check_menu_item_set_active (GTK_CHECK_MENU_ITEM (item), send_html);
}

/**
 * spruce_composer_get_send_html:
 * @composer: A message composer widget
 * 
 * Get the status of the "Send HTML mail" flag.
 * 
 * Return value: The status of the "Send HTML mail" flag.
 **/
gboolean
spruce_composer_get_send_html (SpruceComposer *composer)
{
	g_return_val_if_fail (IS_SPRUCE_COMPOSER (composer), FALSE);
	
	return composer->send_html;
}
