/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* Though it is generally poor practice to use comments in sources as a mode
   of expressing larger thoughts, this is clearly an exception.  Date parsing
   within mail clients is one of the trickiest bits to deal with, due to the
   large number of broken mailers out there and the large number of ways they
   can be broken.  The date parsing code in Spruce (that is, the code which
   parses the date information in messages -received- by Spruce) is designed
   in such a manner as that it covers not only messages with RFC822/1123 
   compliant date fields, but also attempts to deal with non-compliant
   messages.  Persons reviewing this code should keep this detail in mind.

   Spruce should at all times properly process RFC822/RFC1123 compliant
   messages correctly; it will deal with non-compliant messages as best as
   possible.  */

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include "date.h"
#include "parse.h"

static char *days[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
static char *months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", 
			  "Aug", "Sep", "Oct", "Nov", "Dec" };

gint get_days_in_month (gint month, gint year)
{
	/* Return the number of days in a month for a particular year;
	   return 0 on error (unknown month) */
	
	switch (month) {
	case 1: case 3: case 5: case 7: case 8: case 10: case 12:
		return 31;
	case 4: case 6: case 9: case 11:
		return 30;
	case 2:
		if ((year % 4 == 0 && year % 100 != 0) || year % 400 == 0)
			return 29;
		else
			return 28;
	default:
		return 0;
	}
}

gint weekday_toi (gchar *string)
{
	/* Return a numeric equivalent for the weekday passed to the function;
	   return 0 on error (unknown weekday) */
	
	int i;
	
	g_return_val_if_fail ((string != NULL), 0);
	
	for (i = 0; i < 7; i++)
		if (g_strncasecmp (string, days[i], 2) == 0)
			return i + 1;
	
	return 0;  /* unknown week day */
}

gchar *ito_weekday (int weekday)
{
	/* Return a three-letter representation for a specified day of the 
	   week (1-7); return NULL on error (unknown weekday) */
	
	if ((weekday > 7) || (weekday < 1))
		return NULL;
	
	return (days[weekday - 1]);
}

gint month_toi (gchar *string)
{
	/* Return a numeric equivalent for the month passed to the function;
	   return 0 on error (unknown month) */
	
	int i;
	
	g_return_val_if_fail (string != NULL, 0);
	
	for (i = 0; i < 12; i++)
		if (g_strncasecmp (string, months[i], 3) == 0)
			return i + 1; 
	
	return 0;  /* unknown month */
}

gchar *ito_month (gint month)
{
	/* Return a three-letter representation for a specified month of the 
	   year (1-12); return NULL on error (unknown month) */
	
	if ((month > 12) || (month < 1))
		return NULL;
	
	return (months[month - 1]);
}

gdouble rfc1123_timezone_tod (gchar *timezonestring)
{
	/* Return a double representing the timezone offset from UTC (/100.0)
	   using RFC822 timezone information with modifications listed in
	   RFC1123; return -100.0 on error (invalid/unknown timezone) */
	
	/* Military time zones are specified incorrectly in RFC822 and are
	   ignored, per RFC1123 */
	
	/* RFC822-specified time zones */
	if (g_strcasecmp ("UT", timezonestring) == 0)
		return 0.0;
	else if (g_strcasecmp ("GMT", timezonestring) == 0)
		return 0.0;
	else if (g_strcasecmp ("Z", timezonestring) == 0)
		return 0.0;
	else if (g_strcasecmp ("EST", timezonestring) == 0)
		return -5.0;
	else if (g_strcasecmp ("EDT", timezonestring) == 0)
		return -4.0;
	else if (g_strcasecmp ("CST", timezonestring) == 0)
		return -6.0;
	else if (g_strcasecmp ("CDT", timezonestring) == 0)
		return -5.0;
	else if (g_strcasecmp ("MST", timezonestring) == 0)
		return -7.0;
	else if (g_strcasecmp ("MDT", timezonestring) == 0)
		return -6.0;
	else if (g_strcasecmp ("PST", timezonestring) == 0)
		return -8.0;
	else if (g_strcasecmp ("PDT", timezonestring) == 0)
		return -7.0;
	
	return -100.0;
}

date_t adjust_date_t_for_timezone (date_t date)
{
	/* Based on date_t structure, adjust time units based on
	   date.original_tz_offset */
	
	gint tz_offset_minutes, tz_offset_hours;
	
	tz_offset_minutes = (gint)(date.original_tz_offset * 100) % 100;
	tz_offset_hours = (gint)(date.original_tz_offset / 1);
	
	if (date.original_tz_offset != 0) {
		/* Adjust hours and minutes to UTC */
		date.hour -= tz_offset_hours;
		date.minute -= tz_offset_minutes;
		
		/* adjust seconds to proper range */
		if (date.second > 59) {
			date.minute += (date.second / 60);
			date.second = (date.second % 60);
		}
		
		/* adjust minutes to proper range */
		if (date.minute > 59) {
			date.hour += (date.minute / 60);
			date.minute = (date.minute % 60);
		} else if (date.minute < 0) {
				date.minute = -date.minute;
				date.hour -= (date.minute / 60) - 1;
				date.minute = 60 - (date.minute % 60);
		}
		
		/* adjust hours to the proper range */
		if (date.hour > 23) {
			date.day += (date.hour / 24);
			date.hour = (date.hour % 24);
		} else {
			if (date.hour < 0) {
				date.hour = -date.hour;
				date.day -= (date.hour / 24) - 1;
				date.hour = 24 - date.hour;
			}
		}
		
		/* adjust days to the proper range */
		while (date.day > get_days_in_month (date.month, date.year)) {
			date.day -= get_days_in_month (date.month, date.year);
			date.month++;
			if (date.month > 12) {
				date.year += (date.month / 12);
				date.month = (date.month % 12);
				if (date.month == 0) {
					/* month sanity check */
					date.month = 12;
					date.year -= 1;
				}
			}
		}
		
		while (date.day < 1) {
			date.day += get_days_in_month (date.month, date.year);
			date.month--;
			if (date.month < 1) {
				date.month = -date.month;
				date.year -= (date.month / 12) - 1;
				date.month = 12 - (date.month % 12);
			}
		}
		
		/* adjust months to the proper range */
		if (date.month > 12) {
			date.year += (date.month / 12);
			date.month = (date.month % 12);
			if (date.month == 0) {
				/* month sanity check */
				date.month = 12;
				date.year -= 1;
			}
		} else {
			if (date.month < 1) {
				date.month = -date.month;
				date.year -= (date.month / 12) - 1;
				date.month = 12 - (date.month % 12);
			}
		}
	}

	return date;
}

date_t parse_rfc822_date (GList *tokens)
{
	/* Parse RFC822-compliant dates into a date_t structure; flag
	   date.bestguess if this is NOT an RFC822-compliant date */
	guint i = 0, retval = 0;
	gchar *token;
	date_t date;
	
	memset ((void*)&date, 0, sizeof (date_t));
	g_return_val_if_fail (tokens != NULL, date);
	
	/* Check to see if our first parameter is a day of week (optional) */
	if ((token = g_list_nth_data (tokens, i)) == NULL) {
		date.bestguess = 1;
		return date;
	}
	
	if ((retval = weekday_toi (token))) {
		date.day_of_week = retval;
		i++;
	}
	
	/* Next parameter should be day */
	if ((token = g_list_nth_data (tokens, i)) == NULL) {
		date.bestguess = 1;
		return date;
	}
	
	if (isdigit (*token)) {
		retval = atoi (token);
		date.day = retval;
		i++;
	} else {
		date.bestguess = 1;
		return date;
	}
	
	/* Next parameter should be month */
	if ((token = g_list_nth_data (tokens, i)) == NULL) {
		date.bestguess = 1;
		return date;
	}
	
	if ((retval = month_toi (token))) {
		date.month = retval;
		i++;
	} else {
		date.bestguess = 1;
		return date;
	}
	
	/* Next parameter should be year */
	if ((token = g_list_nth_data (tokens, i)) == NULL) {
		date.bestguess = 1;
		return date;
	}
	
	if (isdigit (*token)) {
		retval = atoi (token);
		if (retval < 100) {
			date.year = retval < 69 ? 2000 + atoi (token) 
						: 1900 + atoi (token);
		} else
			date.year = retval;
	} else {
		date.bestguess = 1;
		return date;
	}
	
	/* Verify that our day is valid for month/year */
	if ((token = g_list_nth_data (tokens, i)) == NULL) {
		date.bestguess = 1;
		return date;
	}
	if ((date.day > 0) && (date.day <= get_days_in_month (date.month, 
							      date.year))) {
		i++;
	} else {
		date.bestguess = 1;
		return date;
	}
	
	/* Next field should be time */
	if ((token = g_list_nth_data (tokens, i)) == NULL) {
		date.bestguess = 1;
		return date;
	}
	
	if (strchr (token, ':') && isdigit (*token)) {
		if (strlen (token) == 8) {
			sscanf (token, "%d:%d:%d", &date.hour, &date.minute, 
				&date.second);
		} else if (strlen (token) == 5) {
			sscanf (token, "%d:%d", &date.hour, &date.minute);
			date.second = 0;
		} else {
			date.bestguess = 1;
			return date;
		}
		
		if ((date.hour < 24) && (date.minute < 60) 
		    && (date.second < 60)) {
			i++;
		} else {
			date.bestguess = 1;
			return date;
		}
	} else {
		date.bestguess = 1;
		return date;
	}
	
	/* Next field should be a time zone indicator (only bother with
	   time zone abbreviations if a numeric is not present) */
	if ((token = g_list_nth_data (tokens, i)) == NULL) {
		date.bestguess = 1;
		return date;
	}
	
	if ((*token == '-') || (*token == '+')) {
		date.original_tz_offset = atoi (token) / 100.0;
	} else {
		if ((date.original_tz_offset = rfc1123_timezone_tod (token)) == -100.0) {
			date.bestguess = 1;
			return date;
		}
	}
	
	/* Adjust date for timezone */
	date = adjust_date_t_for_timezone (date);

	/* Return final conclusion */
	return date;
}

date_t parse_date (gchar *datestr)
{
	GList *tokens;
	date_t date;
	
	/* Split the date string into tokens, removing quotes and parenthesis */
	tokens = g_list_strtok (datestr, " \"()");
	
	/* Per RFC1123, it is our duty to properly handle all RFC822+1123
	   dates; if this isn't an RFC-compliant date, make our best guess */
	date = parse_rfc822_date (tokens);
	if (date.bestguess != 1) {
		GList *token;
		
		token = tokens;
		while (token) {
			g_free (token->data);
			token = token->next;
		}
		g_list_free (tokens);
		return date;
	} else {
		/* This is only hit if the date is not RFC822 compliant */
		GList *token;
		
		date = parse_bestguess_date (tokens);
		token = tokens;
		while (token) {
			g_free (token->data);
			token = token->next;
		}
		g_list_free (tokens);
		return date;
	}
}

date_t parse_bestguess_date (GList *tokens)
{
	guint token_count = 0, i, retval = 0;
	date_t date;
	gchar *token, *ptr;
	
	memset ((void*)&date, 0, sizeof (date_t));
	g_return_val_if_fail (tokens != NULL, date);
	date.bestguess = 1;
	
	token_count = g_list_length (tokens);
	
	for (i = 0; i < token_count; i++) {
		token = g_list_nth_data (tokens, i);
		
		if ((retval = weekday_toi (token))) {
			date.day_of_week = retval;
		} else if ((retval = month_toi (token))) {
			date.month = retval;
		} else if (strlen (token) <= 2) {
			/* this could be a 1 or 2 digit day of the month */
			for (retval = 1, ptr = token; *ptr; ptr++)
				if (*ptr < '0' || *ptr > '9')
					retval = 0;
			
			if (retval && atoi (token) <= 31 && !date.day)
				date.day = atoi (token);
			else
				date.year = atoi (token) < 69 ? 2000 + atoi (token) : 1900 + atoi (token);
		} else if (strlen(token) == 4) {
			/* this could be the year... */
			for (retval = 1, ptr = token; *ptr; ptr++)
				if (*ptr < '0' || *ptr > '9')
					retval = 0;
			
			if (retval)
				date.year = atoi (token);
		} else if (strchr (token, ':')) {
			/* this must be the time: hh:mm:ss */
			sscanf (token, "%d:%d:%d", &date.hour, &date.minute, &date.second);
		} else if (*token == '-' || *token == '+') {
			date.original_tz_offset = atoi (token) / 100.0;
		} else if (strcasecmp (token, "PM") == 0) {
			if (date.hour < 12)
				date.hour += 12;
		}
	}
	
	/* Adjust date for timezone */
	date = adjust_date_t_for_timezone (date);
	
	return date;
}
