/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif
                       
#include "globals.h"
#include "server.h"
#include "accounts.h"

/* lets keep track of open windows... */
gboolean AboutOpen;
gboolean AccountsOpen;
gboolean AddrBookOpen;
gboolean FiltersOpen;
gboolean OptionsOpen;
gboolean PassphraseOpen;

/* file and directory locations */
gchar *spruceconf;                       /* Spruce config file */
gchar *homedir;				 /* user home directory */
gchar *mailboxdir;			 /* directory of mailbox files */
gchar *uid_file;			 /* uid file */
gchar *addressbook_file;		 /* address book file */
gchar *sig_file;			 /* signature file */
static gchar *configfile  = ".spruce/spruce.conf";
static gchar *mailbox_dirname = ".spruce";
static gchar *addressfile = ".spruce/addrbook"; /* addressbook file name */
static gchar *sprucesig = ".spruce/sprucesig";

gchar *default_mailboxes[] = { "Inbox", "Outbox", "Drafts", 
			       "Sent items", "Deleted items", "\0" };


/* generic globals */
gint NumFolders;                         /* number of folders (Inbox, Outbox, etc) */

gchar *AddrSelected;                     /* addr selected by frmAddrBook */
gint first_time;
gint texticons;                          /* use text? icons? or both? */
GList *filters;                          /* Filter list */
GList *mime_parts;

/* Identity */
gchar *SenderAddr;                       /* user's email addr */
gchar *ReplyTo;                          /* user's Reply-To email addr */
gint use_rand_sig;                       /* use spruce's random signature? */
gint auto_attach_sig;                    /* auto-attach the signature? */
gchar *signaturefile;                    /* user's signature file */

/* Accounts */
GList *PopAccounts;
GList *ImapAccounts;
Server SmtpServer;                       /* smtp server */
gint external_fetchmail;                 /* use an external fetchmail program */

gint use_rand_sig;                       /* attach a random signature? */
gint save_outgoing;                      /* save outgoing messages? */
Server SmtpServer;                       /* smtp server */

/* Appearance */
gint main_window_layout;                 /* 0 = wide message; 1 = wide headers */
gchar *mesg_body_font;
gchar *compose_mesg_font;
gchar *unread_mesg_header_font;
gchar *read_mesg_header_font;

/* Compose window dimensions, in pixels */
gint compose_window_width;
gint compose_window_height;

/* Word Wrap */
gint word_wrap_len;                      /* chars per line */
gint tab_equivalent;			 /* tab equivalency to spaces */
gint convert_tabs;			 /* convert tabs to spaces? */
gint detect_pgp_clearsigns;		 /* attempt to detect PGP clearsigns */

/* Misc */
gint save_outgoing;                      /* save outgoing messages? */
gint always_prepend;                     /* always prepend Re: or Fw: ?? */
gint empty_on_exit;                      /* empty "Deleted items" on exit? */
gint cm_delay;                           /* check mail every XX minutes */
gint beep_on_new_mail;                   /* make a beep, when new mail arrived */
gint hdr_trim_level;                     /* 0 = none; 1 = partial; 2 = full */
gint save_sizes;
gint save_passwd;

/* PGP */
#ifdef PGP_PROGRAM
gchar *pgp_identity;                     /* PGP Identity */
gint pgp_sign_always;                    /* pgpv -s ? */
gint encrypt_to_self;                    /* always encrypt to self? */
gint store_passphrase;                   /* store our passphrase? */
gchar *pgp_passphrase;                   /* pgp passphrase */
#endif /* PGP_PROGRAM */

/* Sorting preferences */
gint sort_column;                        /* column to sort */
GtkSortType sort_type;

/* here are our window widgets... */
GtkWidget *frmAbout    = NULL;
GtkWidget *frmAccounts = NULL;
GtkWidget *frmAddrBook = NULL;
GtkWidget *frmFilters  = NULL;
GtkWidget *frmMain     = NULL;
GtkWidget *frmOptions  = NULL;
GtkWidget *frmStatus   = NULL;
GtkWidget *dlgPassphrase = NULL;

void reset_globals ()
{
	AboutOpen      = FALSE;
	AccountsOpen   = FALSE;
	AddrBookOpen   = FALSE;
	FiltersOpen    = FALSE;
	OptionsOpen    = FALSE;
	PassphraseOpen = FALSE;
	
	NumFolders = 0;
	
	use_rand_sig = FALSE;
	auto_attach_sig = FALSE;
	save_outgoing = TRUE;
	always_prepend = FALSE;
	
	SmtpServer.port = 25;
	
	external_fetchmail = FALSE;
	save_passwd = TRUE;
	
	/* Set up our file and directory globals */
	
	if ((homedir = getenv ("HOME")) == NULL) {
	/* Bail out, since nothing will work if we can't find our configs */
		fprintf (stderr, "Spruce: unable to determine path to home directory.\n");
		fprintf (stderr, "Please set the 'HOME' environment variable and re-run Spruce.\n\n");
		exit (1);
	}
	
	spruceconf = g_strconcat (homedir, "/", configfile, NULL);
	uid_file = g_strconcat (homedir, "/.spruce/uid.list", NULL);
	mailboxdir = g_strconcat (homedir, "/", mailbox_dirname, NULL);
	addressbook_file = g_strconcat (homedir, "/", addressfile, NULL);
	sig_file = g_strconcat (homedir, "/", sprucesig, NULL);
	
	first_time = 0;
	
	texticons = TB_ICONS;
	
	hdr_trim_level = 1;
	
	empty_on_exit = TRUE;
	
	word_wrap_len = 76;
	tab_equivalent = 1;	/* Tab is equivalent to one space: old behaviour */
	convert_tabs = 0;
	detect_pgp_clearsigns = 0;
	
	cm_delay = 0;
	
	ReplyTo = NULL;
	
	filters = NULL;
	
#ifdef PGP_PROGRAM
	pgp_sign_always = FALSE;
	encrypt_to_self = TRUE;
	store_passphrase = FALSE;
	pgp_passphrase = NULL;
#endif
	
	mime_parts = NULL;
	mesg_body_font = NULL;
	compose_mesg_font = NULL;
	unread_mesg_header_font = NULL;
	read_mesg_header_font = NULL;
	main_window_layout = 0;
	
	PopAccounts = NULL;
	ImapAccounts = NULL;

	compose_window_width = DEFAULT_COMPOSE_WINDOW_WIDTH;
	compose_window_width = DEFAULT_COMPOSE_WINDOW_HEIGHT;
}

void create_default_mailboxes ()
{
	Mailbox_t *mbox;
	gint i;
	
	mkdir (mailboxdir, 0700); /* 0700 = readable/writable only by owner */ 
	
	i = 0;
	while (*default_mailboxes[i]) {
		mbox = mailbox_create (default_mailboxes[i], MAILBOX_LOCAL);
		if (mbox) {
			g_free (mbox->name);
			g_free (mbox);
		}
		i++;
	}
}

void global_free ()
{
	g_free (spruceconf);
	spruceconf = NULL;
	g_free (SenderAddr);
	SenderAddr = NULL;
	g_free (ReplyTo);
	ReplyTo = NULL;
	g_free (signaturefile);
	signaturefile = NULL;
	g_free (AddrSelected);
	AddrSelected = NULL;
#ifdef PGP_PROGRAM
	g_free (pgp_identity);
	pgp_identity = NULL;
	g_free (pgp_passphrase);
	pgp_passphrase = NULL;
#endif /* PGP_PROGRAM */
	g_free (mesg_body_font);
	mesg_body_font = NULL;
	g_free (compose_mesg_font);
	compose_mesg_font = NULL;
	g_free (unread_mesg_header_font);
	unread_mesg_header_font = NULL;
	g_free (read_mesg_header_font);
	read_mesg_header_font = NULL;
	null_accounts ();
}

