/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  Authors: George Talusan
 *           Jeffrey Stedfast <fejj@helixcode.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include "parse.h"
#include "sort.h"
#include "mesg.h"
#include "index.h"

#include "index_imap.h"
#include "imap.h"

static GHashTable *index_table = NULL;

static gint index_imap_list_sort(gconstpointer a, gconstpointer b)
{
	Index_t *first = (Index_t *) a;
	Index_t *second = (Index_t *) b;

	g_return_val_if_fail(first, 0);
	g_return_val_if_fail(second, 0);

	if (index_get_pos(first) < index_get_pos(second))
		return -1;
	if (index_get_pos(first) < index_get_pos(second))
		return 0;
	return 1;
}

static gboolean index_imap_list_free(gpointer key, gpointer data, gpointer user_data)
{
	GList *index_list = (GList *) data;

	g_free(key);
	g_list_foreach(index_list, index_free, GINT_TO_POINTER(FALSE));
	g_list_free(index_list);

	return TRUE;
}

GList *index_imap_load(const Mailbox_t *mbox)
{
	char *raw, *msg, *from, *subject, *date;
	guint i, size, seen;
	GList *index_list;
	Index_t *index;
	
	g_return_val_if_fail(mbox, NULL);
	
	if (!(index_list = g_hash_table_lookup (index_table, mbox->name))) {
		size = index_imap_get_size (mbox);
		
		for (i = 1; i <= size; i++) {
			index = index_new ();
			raw = imap_fetch (mbox->imap->socket, i, "BODY.PEEK[HEADER]", &seen);
			
			g_return_val_if_fail (raw, NULL);
			
			strip (raw, '\r');
			msg = g_strconcat (raw, "\n\n", 0);
			g_free (raw);
			
			from = mesg_get_header_field (msg, "From:");
			subject = mesg_get_header_field (msg, "Subject:");
			date = mesg_get_header_field (msg, "Date:");
			
			index_set_pos (index, i);
			index_set_seen (index, seen);
			index_set_sender (index, from);
			index_set_subject (index, subject);
			index_set_date (index, date);
			
			g_free (from);
			g_free (subject);
			g_free (date);
			
			index_list = g_list_insert_sorted (index_list, index, index_imap_list_sort);
			
			g_free (msg);
		}
		
		g_hash_table_insert (index_table, mbox->name, index_list);
	}
	
	return index_list;
}

void index_imap_commit(void)
{
	;
}

void index_imap_free(void)
{
	if (index_table) {
		g_hash_table_foreach_remove(index_table, index_imap_list_free, NULL);
		g_hash_table_destroy(index_table);
	}
}

void index_imap_init(void)
{
	if (index_table) {
		index_imap_commit();
		index_imap_free();
	}

	index_table = g_hash_table_new(g_str_hash, g_str_equal);
}

void index_imap_create(const Mailbox_t *mbox)
{
	if (imap_mailbox_create (mbox->imap->socket, mbox->imap->namespace, mbox->name) == ERROR)
		return;

	/* we gotta load it here I think, been a while since I last worked with this stuff */
}

Index_t *index_imap_add_msg(const Mailbox_t *mbox, gchar *msg)
{
	return NULL;
}

gchar *index_imap_get_location(const Mailbox_t *mbox)
{
	return NULL;
}

Index_t *index_imap_get_index(const Mailbox_t *mbox, guint i)
{
	GList *index_list;

	g_return_val_if_fail(mbox, NULL);

	index_list = index_imap_load(mbox);
	return g_list_nth_data(index_list, i - 1);
}

gchar *index_imap_get_msg(const Mailbox_t *mbox, Index_t *index)
{
	gchar *msg;
	gint seen;

	g_return_val_if_fail(mbox, NULL);
	g_return_val_if_fail(index, NULL);

	msg = index_get_msg (index);

	if (msg)
		return msg;

	msg = imap_fetch(mbox->imap->socket, index_get_pos(index), "BODY[]", &seen);
	index_set_size(index, strlen(msg));
	index_set_seen(index, seen);
	index_set_msg(index, msg);

	return msg;
}

gboolean index_imap_del_msgs(const Mailbox_t *mbox, GList *unsorted_list)
{
	GList *sorted_list;

	g_return_val_if_fail(mbox, FALSE);
	g_return_val_if_fail(unsorted_list, FALSE);

	sorted_list = g_list_sort (unsorted_list, gint_compare);

	return imap_delete(mbox->imap, sorted_list);
}

guint index_imap_get_size(const Mailbox_t *mbox)
{
	gpointer orig_key = NULL;
	gpointer value = NULL;

	g_return_val_if_fail(mbox, 0);
   
	if (g_hash_table_lookup_extended(index_table, mbox->name, orig_key, value)) {
		g_hash_table_remove(index_table, mbox->name);
		g_free(orig_key);
	}

	return imap_select_mailbox(mbox->imap->socket, mbox->imap->namespace, mbox->name);
}



