/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <utime.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>

#include "mmap.h"
#include "move.h"
#include "parse.h"
#include "sort.h"
#include "mesg.h"
#include "index.h"

#include "index_local.h"

extern gchar *mailboxdir;

#define INDEX_RECORD_FORMAT "%u\n%u\n%u\n%s\n%s\n%s\n"
#define PAGESIZE getpagesize ()

#define PAGE_ALIGN_UP(x,y) (((((x) + (y)) - 1) / (y)) * (y))
#define PAGE_ALIGN_DOWN(x,y) ((((x) - 1) / (y)) * (y))

#ifndef MAP_FAILED
#define MAP_FAILED (-1)
#endif

#define d(x) x

static GHashTable *index_table = NULL;
static gboolean index_local_elem_save_error = FALSE;


/* private prototypes */
static void index_local_elem_save (gpointer data, gpointer user_data);
static void index_local_list_save (gpointer key, gpointer data, gpointer user_data);
static gboolean index_local_list_free (gpointer key, gpointer data, gpointer user_data);
static gboolean index_local_sanity_check (gchar *mailbox_file, gchar *index_file);
static void index_local_commit_internal (void);


/* saves individual elements to the filedes held in user_data */
static void index_local_elem_save (gpointer data, gpointer user_data)
{
	Index_t *index = (Index_t *) data;
	gint fd = GPOINTER_TO_INT (user_data);
	gchar *buf, *sender, *subject, *date;
	guint32 len;
	
	sender = index_get_sender (index);
	subject = index_get_subject (index);
	date = index_get_date (index);
	
	buf = g_strdup_printf (INDEX_RECORD_FORMAT,
			      index_get_pos (index), index_get_size (index), index_get_seen (index),
			      sender ? sender : "(null)",
			      subject ? subject : "(null)",
			      date ? date : "(null)");
	len = strlen (buf);
	
	if (write (fd, buf, len) != len) {
		d(perror ("index_local_elem_save::write ()"));
		index_local_elem_save_error = TRUE;
	}
	g_free (buf);
	
	/* free up some memory */
	index_unset_msg (index);
}

/* saves an entire list in data to the file held in key */
static void index_local_list_save (gpointer key, gpointer data, gpointer user_data)
{
	gchar *mbox_name = g_strdup (key);
	GList *index_list = (GList *) data;
	gchar *index_file, *mailbox_file;
	struct stat stbuf;
	struct utimbuf sanity;
	gint fd;
	
	g_return_if_fail (mbox_name);
	g_return_if_fail (*mbox_name);
	
	mailbox_file = g_strconcat (mailboxdir, "/", mbox_name, ".mbx", NULL);
	index_file = g_strconcat (mailboxdir, "/", mbox_name, ".idx", NULL);
	
	g_free (mbox_name);
	
	/* check to make sure this is a valid mailbox file */
	if (stat (mailbox_file, &stbuf) == -1) {
		g_free (mailbox_file);
		g_free (index_file);
		return;
	}
	
	if (index_list) {
		if ((fd = open (index_file, O_WRONLY | O_CREAT | O_TRUNC, 0600)) == -1) {
			d(perror ("index_local_list_save::open ()"));
			g_free (mailbox_file);
			g_free (index_file);
			return;
		}
		
		index_local_elem_save_error = FALSE;
		g_list_foreach (index_list, index_local_elem_save, GINT_TO_POINTER (fd));
		close (fd);
	} else {
		index_local_elem_save_error = TRUE;
	}
	
	/* if there was an error creating the file, it's corrupt so remove it */
	if (index_local_elem_save_error) {
		remove (index_file);
	} else {
		/* sync times so our sanity check will pass next time index is loaded */
		if (stat (mailbox_file, &stbuf) == -1) {
			remove (index_file);
		} else {
			sanity.actime = stbuf.st_atime;
			sanity.modtime = stbuf.st_mtime;
			
			utime (index_file, &sanity);
		}
	}
	
	g_free (index_file);
	g_free (mailbox_file);
}

/* frees an entire list in data */
static gboolean index_local_list_free (gpointer key, gpointer data, gpointer user_data)
{
	GList *index_list = (GList *) data;
	
	g_free (key);
	g_list_foreach (index_list, index_free, GINT_TO_POINTER (FALSE));
	g_list_free (index_list);
	
	return TRUE;
}

static gboolean index_local_sanity_check (gchar *mailbox_file, gchar *index_file)
{
	/* make sure mailbox/index files are in sync */
	struct stat mailbox_stat, index_stat;
	
	if ((stat (mailbox_file, &mailbox_stat) == -1) || (stat (index_file, &index_stat) == -1))
		return FALSE;
	
	if (mailbox_stat.st_mtime != index_stat.st_mtime
	    || mailbox_stat.st_atime != index_stat.st_atime)
		return FALSE;
	
	return TRUE;
}

GList *index_local_load (const Mailbox_t *mbox)
{
	GList *index_list = NULL;
	gchar *mailbox_file;
	gchar *index_file;
	gchar *ptr, *mmap_ptr;
	gint fd;
	struct stat st;
	
	g_return_val_if_fail (mbox, NULL);
	
	mailbox_file = index_local_get_location (mbox);
	if (!mailbox_file)
		return NULL;
	
	index_file = index_local_get_idx_location (mbox);
	if (!index_file) {
		g_free (mailbox_file);
		return NULL;
	}
	
	index_list = g_hash_table_lookup (index_table, mbox->name);
	
	if (!index_list) {
		fd = open (index_file, O_RDONLY);
		
		if (fd != -1 && index_local_sanity_check (mailbox_file, index_file)) {
			if (fstat (fd, &st)) {
				d(perror ("index_local_load::fstat ()"));
				g_free (mailbox_file);
				g_free (index_file);
				
				return NULL;
			}
			
			ptr = mmap_ptr = (gchar *) g_mmap (0, st.st_size + PAGESIZE, PROT_READ, MAP_PRIVATE, fd, 0);
			
			if (ptr == (gchar *) MAP_FAILED) {
				d(perror ("index_local_load::g_mmap ()"));
				close (fd);
				g_free (mailbox_file);
				g_free (index_file);
				
				return NULL;
			}
			
			while (ptr < st.st_size + mmap_ptr) {
				guint32 pos, size;
				gboolean error_flag, seen;
				
				pos = strtoul (ptr, &ptr, 10);
				error_flag = (pos == ULONG_MAX && errno == ERANGE);
				
				if (!error_flag) {
					size = strtoul (ptr, &ptr, 10);
					error_flag = (size == ULONG_MAX && errno == ERANGE);
				}
				
				if (!error_flag) {
					seen = (gboolean) strtoul (ptr, &ptr, 10);
					error_flag = (seen == ULONG_MAX && errno == ERANGE);
				}
				
				if (!error_flag) {
					gchar *end;
					gchar *extract;
					Index_t *index;
					
					index = index_new ();
					index_set_pos (index, pos);
					index_set_size (index, size);
					index_set_seen (index, seen);
					
					ptr = strchr (ptr, '\n') + 1;
					end = strchr (ptr, '\n'); 
					extract = g_strndup (ptr, end - ptr);
					index_set_sender (index, extract);
					g_free (extract);
					
					ptr = end + 1;
					end = strchr (ptr, '\n');
					extract = g_strndup (ptr, end - ptr);
					index_set_subject (index, extract);
					g_free (extract);
					
					ptr = end + 1;
					end = strchr (ptr, '\n');
					extract = g_strndup (ptr, end - ptr);
					index_set_date (index, extract);
					g_free (extract);
					
					index_list = g_list_append (index_list, index);
					ptr = end + 1;
				} else {
					g_print ("index file for %s is corrupt\n", mailbox_file);
					g_munmap (mmap_ptr, st.st_size);
					close (fd);
					g_free (mailbox_file);
					g_free (index_file);
					
					return NULL;
				}
			}
			
			g_hash_table_insert (index_table, mbox->name, index_list);
			g_munmap (mmap_ptr, st.st_size);
			close (fd);
		} else {
			if (fd != -1) {
				close (fd);
				d(fprintf (stderr, "index_local_load: %s.idx did not pass sanity check...recreating.\n",
					mbox->name));
			}
			index_local_create (mbox);
			index_list = g_hash_table_lookup (index_table, mbox->name);
		}
	}
	
	g_free (mailbox_file);
	g_free (index_file);
	
	return index_list;
}

/* commits all indexes to disk */
static void index_local_commit_internal (void)
{
	/* had i been thinking when i designed the indexes, we'd
	 * be doing this only for dirty indexes */
	
	g_hash_table_foreach (index_table, index_local_list_save, NULL);
}

void index_local_commit (void)
{
	index_local_commit_internal ();
}

/* free all the indexes from memory */
void index_local_free (void)
{
	if (index_table) {
		g_hash_table_foreach_remove (index_table, index_local_list_free, NULL);
		g_hash_table_destroy (index_table);
		index_table = NULL;
	}
}

#if 0
static gboolean index_local_mutex_free (gpointer key, gpointer data, gpointer user_data)
{
	g_free (key);
	g_free (data);
	return TRUE;
}

static void index_local_free_mutexes (void)
{
	if (mutex_table) {
		g_hash_table_foreach_remove (mutex_table, index_local_mutex_free, NULL);
		g_hash_table_destroy (mutex_table);
		mutex_table = NULL;
	}
}
#endif

/* create indexing data structure */
void index_local_init (void)
{
	gboolean condition = FALSE;
	
	condition = (index_table && 1);
	
	if (condition) {
		index_local_commit_internal ();
		index_local_free ();
	}
	
	/* prepare the hash table for demand loading of index files */
	index_table = g_hash_table_new (g_str_hash, g_str_equal);
}

/* parses a given mailbox file and creates an index file for it */
void index_local_create (const Mailbox_t *mbox)
{
	struct stat st;
	gchar *p, *q, *r, *end;
	gint fd;
	gchar *mailbox_file;
	gpointer orig_key = NULL, value = NULL;
	GList *index_list = NULL;
	
	g_return_if_fail (mbox);
	
	mailbox_file = index_local_get_location (mbox);
	g_return_if_fail (mailbox_file);
	
	if (stat (mailbox_file, &st) == -1) {
		d(perror ("index_local_create::stat"));
		g_free (mailbox_file);
		
		return;
	}
	
	if ((fd = open (mailbox_file, O_RDONLY)) == -1) {
		d(perror ("index_local_create::open ()"));
		creat (mailbox_file, S_IRUSR | S_IWUSR);
		g_free (mailbox_file);
		
		return;
	}
	
	if (!st.st_size) {
		/* g_print ("index_local_create: mailbox %s is empty\n", mailbox_file); */
		close (fd);
		g_free (mailbox_file);
		
		return;
	}
	
	p = q = (gchar *) g_mmap (0, st.st_size, PROT_READ, MAP_PRIVATE, fd, 0);
	
	if (p == (gchar *) MAP_FAILED) {
		d(perror ("index_local_create::g_mmap ()"));
		close (fd);
		g_free (mailbox_file);
		
		return;
	}
	
	if (g_hash_table_lookup_extended (index_table, mbox->name, orig_key, value)) {
		g_hash_table_remove (index_table, orig_key);
		index_local_list_free (orig_key, value, NULL);
	}
	
	end = p + st.st_size;
	/* p and r point to the beginning and end of an email message */
	while ((r = strstrbound (p, "\n.\r\n", end))) {
		Index_t *index = index_new ();
		char *msg, *from, *subject, *date;
		
		r += 4; /* skip over \n.\r\n */
		
		msg = g_strndup (p, (guint32) (r - p));
		
		from = mesg_get_header_field (msg, "From:");
		subject = mesg_get_header_field (msg, "Subject:");
		date = mesg_get_header_field (msg, "Date:");
		
		index_set_pos (index, (guint32) (p - q));  /* get physical position */
		index_set_size (index, (guint32) (r - p)); /* get physical size */
		index_set_seen (index, FALSE);
		index_set_sender (index, from);
		index_set_subject (index, subject);
		index_set_date (index, date);
		
		g_free (from);
		g_free (subject);
		g_free (date);
		
		index_list = g_list_append (index_list, index);
		
		p = r;
		g_free (msg);
	}
	
	g_hash_table_insert (index_table, mbox->name, index_list);
	index_local_commit_internal ();
		
	g_munmap (q, st.st_size);
	close (fd);
	
	g_free (mailbox_file);
}

/* adds a message to an index, and appends it to the mailbox */
Index_t *index_local_add_msg (const Mailbox_t *mbox, const gchar *msg)
{
	GList *index_list;
	struct stat st;
	gint fd;
	gchar *mailbox_file;
	gchar *index_file;
	Index_t *index = NULL;
	gboolean free_key = TRUE;
	
	g_return_val_if_fail (mbox, NULL);
	
	mailbox_file = index_local_get_location (mbox);
	if (!mailbox_file)
		return NULL;
	
	index_file = index_local_get_idx_location (mbox);
	if (!index_file) {
		g_free (mailbox_file);
		return NULL;
	}
	
	if (stat (index_file, &st) == -1) {
		/* file doesn't exist */
		if (errno != ENOENT) {
			g_free (mailbox_file);
			g_free (index_file);
			return NULL;
		}
		index_local_create (mbox);
		memset (&st, 0, sizeof (struct stat));
	}
	
	if (stat (mailbox_file, &st) == -1) {
		/* mailbox doesn't exist, lets try creating it before we return an error */
		if ((fd = open (mailbox_file, O_WRONLY | O_CREAT, 0600)) == -1) {
			d(perror ("index_local_add_msg::stat ()"));
			g_free (mailbox_file);
			g_free (index_file);
			
			return NULL;
		}
		close (fd);
	}
	
	free_key = (gboolean) !index_local_load (mbox);
	
	if ((fd = open (index_file, O_WRONLY | O_CREAT /* | O_APPEND ?? */, 0600)) != -1) {
		gchar *mesg;
		
		if (!strstr (msg, "\n.\r\n"))
			mesg = g_strconcat (msg, "\n.\r\n", NULL);
		else
			mesg = g_strdup (msg);
		
		index = index_new ();
		index_set_pos (index, st.st_size);
		index_set_size (index, strlen (mesg));
		index_set_seen (index, FALSE);
		index_set_sender (index, mesg_get_header_field (mesg, "\nFrom:"));
		index_set_subject (index, mesg_get_header_field (mesg, "\nSubject:"));
		/* Date is the first field in a trimmed header message, so must
		   not be preceeded by \n in mesg_get_header_field() call */
		index_set_date (index, mesg_get_header_field (mesg, "Date:"));
		
		index_list = g_hash_table_lookup (index_table, mbox->name);
		index_list = g_list_append (index_list, index);
		g_hash_table_insert (index_table, mbox->name, index_list);
		
		lseek (fd, 0, SEEK_END);  /* seek to the end so we append */
		index_local_elem_save (index, GINT_TO_POINTER (fd));
		close (fd);
		
		/* now store the message in the mailbox */
		if ((fd = open (mailbox_file, O_WRONLY | O_CREAT, 0600)) != -1) {
			guint32 len;
			
			lseek (fd, 0, SEEK_END);  /* seek to the end so we append */
			len = strlen (mesg);
			
			if (write (fd, mesg, len) != len) {
				d(perror ("index_local_add_msg::write ()"));
				d(fprintf (stderr, "index_local_add_msg::%s::Some data may have been lost.\n", mbox->name));
			}
			close (fd);
		} else {
			d(perror ("index_local_add_msg::open ()"));
			d(fprintf (stderr, "index_local_add_msg::Undoing changes made to index...\n"));
			index_local_create (mbox);
		}
		
		g_free (mesg);
	}
	
	g_free (mailbox_file);
	
	g_free (index_file);
	
	return index;
}

gchar *index_local_get_location (const Mailbox_t *mbox)
{
	gchar *location;
	
	g_return_val_if_fail (mbox, NULL);
	
	location = g_strconcat (mailboxdir, "/", mbox->name, ".mbx", NULL);
	
	return location;
}

gchar *index_local_get_idx_location (const Mailbox_t *mbox)
{
	gchar *location;
	
	g_return_val_if_fail (mbox, NULL);
	
	location = g_strconcat (mailboxdir, "/", mbox->name, ".idx", NULL);
	
	return location;
}

/* fetches a message by a given index from a mailbox */
gchar *index_local_get_msg (const Mailbox_t *mbox, Index_t *index)
{
	gchar *mailbox_file;
	gint fd;
	gchar *p, *msg;
	guint32 ofs, end_ofs;
	struct stat st;
	
	g_return_val_if_fail (mbox, NULL);
	
	msg = index_get_msg(index);
	
	if (msg)
		return msg;
	
	mailbox_file = index_local_get_location (mbox);
	g_return_val_if_fail (mailbox_file, NULL);
	
	if (stat (mailbox_file, &st) == -1) {
		d(perror ("index_local_get_msg::stat ()"));
		g_free (mailbox_file);
		return NULL;
	}
	
	if ((fd = open (mailbox_file, O_RDONLY)) == -1) {
		d(perror ("index_local_get_msg::open ()"));
		g_free (mailbox_file);
		return NULL;
	}
	
	ofs = PAGE_ALIGN_DOWN (index_get_pos (index), PAGESIZE);
	end_ofs = PAGE_ALIGN_UP (index_get_pos (index) + index_get_size (index), PAGESIZE);
	
	if (ofs > end_ofs)
		ofs = 0;
	
	p = (gchar *) g_mmap (0, end_ofs - ofs + PAGESIZE, PROT_READ, MAP_PRIVATE, fd, ofs);
	if (p == (gchar *) MAP_FAILED) {
		d(perror ("index_local_get_msg::g_mmap ()"));
		close (fd);
		g_free (mailbox_file);
		return NULL;
	}
	
	/* seek the message, and forget the "\n.\r\n" */
	msg = g_strndup (p + index_get_pos (index) - ofs, index_get_size (index) - 4);
	strip (msg, '\r');
	index_set_msg (index, msg);
	
	g_munmap (p, end_ofs - ofs + PAGESIZE);
	close (fd);
	g_free (mailbox_file);
	
	return msg;
}

Index_t *index_local_get_index (const Mailbox_t *mbox, guint i)
{
	GList *index_list = index_local_load (mbox);
	Index_t *index = g_list_nth_data (index_list, i - 1);
	
	return index;
}

gboolean index_local_del_msgs (const Mailbox_t *mbox, GList *unsorted_list)
{
	GList *sorted_list;
	GList *index_list;
	GList *del_list;
	gchar *tmp_box_file, *mailbox_file;
	guint i, j, list_len, total_del_size;
	char *template;
	Mailbox_t tmp_box;
	gint fd;
	
	g_return_val_if_fail (unsorted_list, FALSE);
	
	sorted_list = g_list_sort (unsorted_list, gint_compare);
	del_list = NULL;
	list_len = g_list_length (sorted_list);
	total_del_size = 0;
	
	/* generate a somewhat random file name */
	template = g_strdup_printf ("spruce-local-spool-%d-%d", (int) time (NULL), getpid ());
	if (!template)
		return FALSE;
	
	tmp_box.name = template;
	tmp_box.type = MAILBOX_LOCAL;
	
	tmp_box_file = index_local_get_location (&tmp_box);
	mailbox_file = index_local_get_location (mbox);
	
	/* we don't need this anymore */
	g_free (template);
	
	if ((fd = open (tmp_box_file, O_WRONLY | O_CREAT | O_TRUNC, 0600)) == -1) {
		d(perror ("index_local_del_msgs::open ()"));
		g_free (tmp_box_file);
		g_free (mailbox_file);
		
		return FALSE;
	}
	
	for (i = 0, j = 1; i < list_len; i++, j++) {
		/* get the id # to delete */
		gint msg_num = (gint) g_list_nth_data (sorted_list, i);
		Index_t *del_index = index_local_get_index (mbox, msg_num);
		
		/* remember each index to be deleted */
		del_list = g_list_append (del_list, del_index);
		
		while (j < msg_num) {
			Index_t *index = index_local_get_index (mbox, j);
			gchar *msg = index_local_get_msg (mbox, index);
			guint32 len = strlen (msg);
			
			if (write (fd, msg, len) != len || write (fd, "\n.\r\n", 4) != 4) {
				d(perror ("index_local_del_msgs::write ()"));
				close (fd);
				remove (tmp_box_file);
				g_free (tmp_box_file);
				g_free (mailbox_file);
				
				return FALSE;
			}
			
			/* recalculate index offsets.  add 4 to account for \n.\r\n */
			index_set_pos (index, index_get_pos (index) - total_del_size);
			j++;
		}
		total_del_size += index_get_size (del_index);
	}
	
	for (i = j; i <= index_local_get_size (mbox); i++) {
		Index_t *index = index_local_get_index (mbox, i);
		gchar *msg = index_local_get_msg (mbox, index);
		guint32 len = strlen (msg);
		
		if (write (fd, msg, len) != len || write (fd, "\n.\r\n", 4) != 4) {
			d(perror ("index_local_del_msgs::write ()"));
			close (fd);
			remove (tmp_box_file);
			g_free (tmp_box_file);
			g_free (mailbox_file);
			
			return FALSE;
		}
		
		/* let's go easy on memory */
		index_unset_msg (index);
		
		index_set_pos (index, index_get_pos (index) - total_del_size);
	}
	close (fd);
	
	/* now purge the index list of the deleted indices */
	index_list = index_local_load (mbox);
	
	while (g_list_length (del_list)) {
		Index_t *index = g_list_nth_data (del_list, 0);
		
		/* remove the index from the list */
		index_list = g_list_remove (index_list, index);
		
		/* remove the index from the temporary list */
		del_list = g_list_remove (del_list, index);
		index_free (index, GINT_TO_POINTER (FALSE));
	}
	
	if (move (tmp_box_file, mailbox_file) == -1) {
		d(perror ("index_local_del_msgs::move ()"));
		remove (tmp_box_file);
		g_free (tmp_box_file);
		g_free (mailbox_file);
		
		return FALSE;
	}
	g_free (tmp_box_file);
	
	g_hash_table_insert (index_table, mbox->name, index_list);
	index_local_commit_internal ();
	
	g_free (mailbox_file);
	
	return TRUE;
}

guint index_local_get_size (const Mailbox_t *mbox)
{
	GList *index_list = index_local_load (mbox);
	
	return g_list_length (index_list);
}

guint index_local_get_unread (const Mailbox_t *mbox)
{
	GList *index_list = index_local_load (mbox), *tmp;
	int unread = 0;

	tmp = index_list;
	while (tmp != NULL) {
		if (index_get_seen (tmp->data) == 0) {
			unread++;
		}
		tmp = tmp->next;
	}

	return unread;
}
