/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  Authors: George Talusan
 *           Jeffrey Stedfast <fejj@helixcode.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <stdio.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <utime.h>
#include <fcntl.h>
#define __USE_GNU
#include <unistd.h>

#include "mmap.h"
#include "move.h"
#include "parse.h"
#include "mesg.h"
#include "sort.h"
#include "index.h"

#include "index_spool.h"

#define INDEX_RECORD_FORMAT "%u\n%u\n%u\n%s\n%s\n%s\n"
#define PAGESIZE getpagesize ()

#define PAGE_ALIGN_UP(x,y) (((((x) + (y)) - 1) / (y)) * (y))
#define PAGE_ALIGN_DOWN(x,y) ((((x) - 1) / (y)) * (y))

#define d(x) x

extern gchar *mailboxdir;

static GHashTable *index_table = NULL;
static gboolean index_spool_elem_save_error = FALSE;

/* private prototypes */
static gchar *index_spool_get_index_file (const gchar *mailbox);
static void index_spool_elem_save (gpointer data, gpointer user_data);
static void index_spool_list_save (gpointer key, gpointer data, gpointer user_data);
static gboolean index_spool_list_free (gpointer key, gpointer data, gpointer user_data);
static gboolean index_spool_sanity_check (gchar *mailbox_file, gchar *index_file);


static gchar *index_spool_get_index_file (const gchar *mailbox)
{
	gchar **i, **v;
	gchar *s;
	
	g_return_val_if_fail (mailbox, NULL);
	
	v = g_strsplit (mailbox, "/", 0);
	
	for (i = v; *i; i++); /* nasty */
	i--;
	s = g_strconcat (mailboxdir, "/", *i, ".idx", NULL);
	g_strfreev (v);
	
	return s;
}

static void index_spool_elem_save (gpointer data, gpointer user_data)
{
	Index_t *index = (Index_t *) data;
	gint fd = GPOINTER_TO_INT (user_data);
	gchar *buf, *sender, *subject, *date;
	guint32 len;
	
	sender = index_get_sender (index);
	subject = index_get_subject (index);
	date = index_get_date (index);
	
	buf = g_strdup_printf(INDEX_RECORD_FORMAT,
			      index_get_pos (index), index_get_size (index), index_get_seen (index),
			      sender ? sender : "(null)",
			      subject ? subject : "(null)",
			      date ? date : "(null)");
	len = strlen (buf);
	
	if (write (fd, buf, len) != len) {
		d(perror ("index_local_elem_save::write ()"));
		index_spool_elem_save_error = TRUE;
	}
	g_free (buf);
	
	/* free up some memory */
	index_unset_msg (index);
}

static void index_spool_list_save (gpointer key, gpointer data, gpointer user_data)
{
	gchar *mbox_name = g_strdup (key);
	GList *index_list = (GList *) data;
	gchar *index_file, *mailbox_file;
	struct stat stbuf;
	struct utimbuf sanity;
	gint fd;
	
	g_return_if_fail (mbox_name);
	
	mailbox_file = g_strdup (mbox_name);
	index_file = index_spool_get_index_file (mbox_name);
	
	g_free (mbox_name);
	
	if (index_list) {
		if ((fd = open (index_file, O_WRONLY | O_CREAT | O_TRUNC, 0600)) == -1) {
			d(perror ("index_spool_list_save::open ()"));
			g_free (mailbox_file);
			g_free (index_file);
			
			return;
		}
		
		index_spool_elem_save_error = FALSE;
		g_list_foreach (index_list, index_spool_elem_save, GINT_TO_POINTER (fd));
		close (fd);
	} else {
		index_spool_elem_save_error = TRUE;
	}
	
	/* if there was an error creating the file, it's corrupt so remove it */
	if (index_spool_elem_save_error) {
		remove (index_file);
	} else {
		/* sync times so our sanity check will pass next time index is loaded */
		stat (mailbox_file, &stbuf);
		sanity.actime = stbuf.st_atime;
		sanity.modtime = stbuf.st_mtime;
		
		utime (index_file, &sanity);
	}
	
	g_free (index_file);
}

static gboolean index_spool_list_free (gpointer key, gpointer data, gpointer user_data)
{
	GList *index_list = (GList *) data;
	
	g_free (key);
	g_list_foreach (index_list, index_free, GINT_TO_POINTER (FALSE));
	g_list_free (index_list);
	
	return TRUE;
}

static gboolean index_spool_sanity_check (gchar *mailbox_file, gchar *index_file)
{
	/* make sure mailbox/index files are in sync */
	struct stat mailbox_stat, index_stat;
	
	if ( (stat (mailbox_file, &mailbox_stat) == -1) || (stat (index_file, &index_stat) == -1) )
		return FALSE;
	
	if (mailbox_stat.st_mtime != index_stat.st_mtime
	    || mailbox_stat.st_atime != index_stat.st_atime)
		return FALSE;
	
	return TRUE;
}

GList *index_spool_load (const Mailbox_t *mbox)
{
	GList *index_list;
	gchar *mailbox_file;
	gchar *index_file = NULL;
	gchar *ptr, *mmap_ptr;
	gint fd;
	struct stat st;
	
	g_return_val_if_fail (mbox, NULL);
	
	mailbox_file = index_spool_get_location (mbox);
	g_return_val_if_fail (mailbox_file, NULL);
	
	if (!(index_list = g_hash_table_lookup (index_table, mailbox_file))) {
		index_file = index_spool_get_index_file (mbox->name);
		fd = open (index_file, O_RDONLY);
		
		if (fd != -1 && index_spool_sanity_check (mailbox_file, index_file)) {
			if (fstat (fd, &st)) {
				d(perror ("index_spool_load::fstat ()"));
				g_free (mailbox_file);
				g_free (index_file);
				
				return NULL;
			}
			
			ptr = mmap_ptr = (gchar *) g_mmap (0, st.st_size + PAGESIZE, PROT_READ, MAP_PRIVATE, fd, 0);
			
			if (ptr == (gchar *) MAP_FAILED) {
				d(perror ("index_spool_load::g_mmap ()"));
				close (fd);
				g_free (mailbox_file);
				g_free (index_file);
				
				return NULL;
			}
			
			while (ptr < st.st_size + mmap_ptr) {
				guint32 pos, size;
				gboolean error_flag, seen;
				
				pos = strtoul (ptr, &ptr, 10);
				error_flag = (pos == ULONG_MAX && errno == ERANGE);
				
				if (!error_flag) {
					size = strtoul (ptr, &ptr, 10);
					error_flag = (size == ULONG_MAX && errno == ERANGE);
				}
				
				if (!error_flag) {
					seen = (gboolean) strtoul (ptr, &ptr, 10);
					error_flag = (seen == ULONG_MAX && errno == ERANGE);
				}
				
				if (!error_flag) {
					gchar *end;
					gchar *extract;
					Index_t *index;
					
					index = index_new ();
					index_set_pos (index, pos);
					index_set_size (index, size);
					index_set_seen (index, seen);
					
					ptr = strchr (ptr, '\n') + 1;
					end = strchr (ptr, '\n');
					extract = g_strndup (ptr, end - ptr);
					index_set_sender (index, extract);
					g_free (extract);
					
					ptr = end + 1;
					end = strchr (ptr, '\n');
					extract = g_strndup (ptr, end - ptr);
					index_set_subject (index, extract);
					g_free (extract);
					
					ptr = end + 1;
					end = strchr (ptr, '\n');
					extract = g_strndup (ptr, end - ptr);
					index_set_date (index, extract);
					g_free (extract);
					
					index_list = g_list_append (index_list, index);
					ptr = end + 1;
				} else {
					d(fprintf (stderr, "index file for %s is corrupt\n", mailbox_file));
					g_munmap (mmap_ptr, st.st_size);
					close (fd);
					g_free (mailbox_file);
					g_free (index_file);
					
					return NULL;
				}
			}
			g_hash_table_insert (index_table, mbox->name, index_list);
			g_munmap (mmap_ptr, st.st_size);
			close (fd);
		} else {
			if (fd != -1) {
				close (fd);
				d(fprintf (stderr, "index_spool_load: %s did not pass sanity check...recreating.\n",
					   index_file));
			}
			index_spool_create (mbox);
			index_list = g_hash_table_lookup (index_table, mbox->name);
		}
	}
	
	g_free (mailbox_file);
	g_free (index_file);
	
	return index_list;
}

void index_spool_free (void)
{
	if (index_table) {
		g_hash_table_foreach_remove (index_table, index_spool_list_free, NULL);
		g_hash_table_destroy (index_table);
		index_table = NULL;
	}
}

void index_spool_init (void)
{
	if (index_table) {
		index_spool_commit ();
		index_spool_free ();
	}
	
	index_table = g_hash_table_new(g_str_hash, g_str_equal);
}

void index_spool_commit (void)
{
	g_hash_table_foreach (index_table, index_spool_list_save, NULL);
}

void index_spool_create (const Mailbox_t *mbox)
{
	struct stat st;
	gchar *p, *q, *r, *end;
	gchar *mailbox_file;
	gint fd;
	gpointer orig_key = NULL, value = NULL;
	GList *index_list = NULL;
	
	g_return_if_fail (mbox);
	
	mailbox_file = g_strdup (mbox->name);
	g_return_if_fail (mailbox_file);
	
	if (stat (mailbox_file, &st) == -1) {
		g_free (mailbox_file);
		d(perror ("index_spool_create::stat ()"));
		return;
	}
	
	if ((fd = open (mailbox_file, O_RDONLY)) == -1) {
		g_free (mailbox_file);
		d(perror ("index_spool_create::open ()"));
		return;
	}
	
	if (!st.st_size) {
		g_free (mailbox_file);
		close (fd);
		return;
	}
	
	p = q = (gchar *) g_mmap (0, st.st_size + PAGESIZE, PROT_READ, MAP_PRIVATE, fd, 0);
	
	if (p == (gchar *) MAP_FAILED) {
		g_free (mailbox_file);
		d(perror ("index_spool_create::g_mmap ()"));
		close (fd);
		return;
	}
	
	if (g_hash_table_lookup_extended (index_table, mailbox_file, orig_key, value)) {
		g_hash_table_remove (index_table, orig_key);
		index_spool_list_free (orig_key, value, NULL);
	}
	
	end = p + st.st_size;
	
	/* need to make this a bit more robust.. we shouldn't check Return-Path
	 * since not all messages have it.
	 */
	while (TRUE) {
		Index_t *index = index_new ();
		char *msg, *from, *subject, *date;
		
		r = strstrbound (p, "\nFrom ", end); /* won't match first message because of \n */
		
		if (r == NULL) /* at the last message? */
			r = (char *) (q + st.st_size);
		else
			r++;
		
		msg = g_strndup (p, (guint32) (r - p));
		
		from = mesg_get_header_field (msg, "From:");
		subject = mesg_get_header_field (msg, "Subject:");
		date = mesg_get_header_field (msg, "Date:");
		
		index_set_pos (index, (guint32) (p - q));
		index_set_size (index, (guint32) (r - p));
		index_set_seen (index, FALSE);
		index_set_sender (index, from);
		index_set_subject (index, subject);
		index_set_date (index, date);
		
		g_free (from);
		g_free (subject);
		g_free (date);
		
		index_list = g_list_append (index_list, index);
		
		g_free (msg);
		
		if (r == (gchar *) (q + st.st_size))
			break;
		p = r;
	}
	
	g_hash_table_insert (index_table, mailbox_file, index_list);
	
	index_spool_commit ();
	
	g_munmap (q, st.st_size + PAGESIZE);
	close (fd);
	
	g_free (mailbox_file);
}

Index_t *index_spool_add_msg (const Mailbox_t *mbox, const gchar *msg)
{
	return NULL;
}

gchar *index_spool_get_location (const Mailbox_t *mbox)
{
	return g_strdup (mbox->name);
}

gchar *index_spool_get_msg (const Mailbox_t *mbox, Index_t *index)
{
	gint fd;
	gchar *p, *msg;
	guint32 ofs, end_ofs;
	struct stat st;
	
	msg = index_get_msg (index);
	
	if (msg)
		return msg;
	
	g_return_val_if_fail (mbox, NULL);
	
	if (stat (mbox->name, &st) == -1) {
		d(perror ("index_spool_get_msg::stat ()"));
		return NULL;
	}
	
	if ((fd = open (mbox->name, O_RDONLY)) == -1) {
		d(perror ("index_spool_get_msg::open ()"));
		return NULL;
	}
	
	ofs = PAGE_ALIGN_DOWN (index_get_pos (index), PAGESIZE);
	end_ofs = PAGE_ALIGN_UP (index_get_pos (index) + index_get_size (index), PAGESIZE);
	
	if (ofs > end_ofs)
		ofs = 0;
	
	p = (gchar *) g_mmap (0, end_ofs - ofs + PAGESIZE, PROT_READ,
			      MAP_PRIVATE, fd, ofs);
	if (p == (gchar *) MAP_FAILED) {
		d(perror ("index_spool_get_msg::g_mmap ()"));
		close (fd);
		return NULL;
	}
	
	msg = g_strndup (p + index_get_pos (index) - ofs, index_get_size (index));
	strip (msg, '\r');
	index_set_msg (index, msg);
	g_munmap (p, end_ofs - ofs + PAGESIZE);
	close (fd);
	
	return msg;
}

Index_t *index_spool_get_index (const Mailbox_t *mbox, guint i)
{
	GList *index_list = index_spool_load (mbox);
	Index_t *index = g_list_nth_data (index_list, i - 1);
	
	return index;
}

gboolean index_spool_del_msgs (const Mailbox_t *mbox, GList *unsorted_list)
{
	GList *sorted_list;
	GList *index_list;
	GList *del_list;
	gchar *tmp_box_file, *mailbox_file;
	guint i, j, list_len, total_del_size;
	char *template;
	gint fd;
	
	g_return_val_if_fail (unsorted_list, FALSE);
	
	sorted_list = g_list_sort (unsorted_list, gint_compare);
	del_list = NULL;
	list_len = g_list_length (sorted_list);
	total_del_size = 0;
	
	/* generate a somewhat random file name */
	template = g_strdup_printf ("spruce-mbox-spool-%d-%d", (int) time (NULL), getpid ());
	if (!template)
		return FALSE;
	
	tmp_box_file = g_strjoin (mailboxdir, template, NULL);
	mailbox_file = index_spool_get_location (mbox);
	
	/* we don't need this anymore */
	g_free (template);
	
	if ((fd = open (tmp_box_file, O_WRONLY | O_CREAT | O_TRUNC, 0600)) == -1) {
		d(perror ("index_local_del_msgs::open ()"));
		g_free (tmp_box_file);
		g_free (mailbox_file);
		return FALSE;
	}
	
	for (i = 0, j = 1; i < list_len; i++, j++) {
		/* get the id # to delete */
		gint msg_num = (gint) g_list_nth_data (sorted_list, i);
		Index_t *del_index = index_spool_get_index (mbox, msg_num);
		
		/* remember each index to be deleted */
		del_list = g_list_append (del_list, del_index);
		
		while (j < msg_num) {
			Index_t *index = index_spool_get_index (mbox, j);
			gchar *msg = index_spool_get_msg (mbox, index);
			guint32 len = strlen (msg);
			
			if (write (fd, msg, len) != len) {
				d(perror ("index_local_del_msgs::write ()"));
				close (fd);
				remove (tmp_box_file);
				g_free (tmp_box_file);
				g_free (mailbox_file);
				return FALSE;
			}
			
			/* recalculate index offsets.  add 4 to account for \n.\r\n */
			index_set_pos (index, index_get_pos (index) - total_del_size);
			j++;
		}
		total_del_size += index_get_size (del_index);
	}
	
	for (i = j; i <= index_spool_get_size (mbox); i++) {
		Index_t *index = index_spool_get_index (mbox, i);
		gchar *msg = index_spool_get_msg (mbox, index);
		guint32 len = strlen (msg);
		
		if (write (fd, msg, len) != len) {
			d(perror ("index_local_del_msgs::write ()"));
			close (fd);
			remove (tmp_box_file);
			g_free (tmp_box_file);
			g_free (mailbox_file);
			
			return FALSE;
		}
		
		/* let's go easy on memory */
		index_unset_msg (index);
		
		index_set_pos (index, index_get_pos (index) - total_del_size);
	}
	close (fd);
	
	/* now purge the index list of the deleted indices */
	index_list = index_spool_load (mbox);
	
	while (g_list_length (del_list)) {
		Index_t *index = g_list_nth_data (del_list, 0);
		
		/* remove the index from the list */
		index_list = g_list_remove (index_list, index);
		
		/* remove the index from the temporary list */
		del_list = g_list_remove (del_list, index);
		index_free (index, GINT_TO_POINTER (FALSE));
	}
	
	if (move (tmp_box_file, mailbox_file) == -1) {
		d(perror ("index_local_del_msgs::move ()"));
		remove (tmp_box_file);
		g_free (tmp_box_file);
		g_free (mailbox_file);
		
		return FALSE;
	}
	
	g_hash_table_insert (index_table, mailbox_file, index_list);
	index_spool_commit ();
	
	g_free (tmp_box_file);
	g_free (mailbox_file);
	
	return TRUE;
}

guint index_spool_get_size (const Mailbox_t *mbox)
{
	GList *index_list = index_spool_load (mbox);
	
	return g_list_length (index_list);
}
