/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  Authors: George Talusan
 *           Jeffrey Stedfast <fejj@helixcode.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/* TODO: mark {all, selected} {un}read. */

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/mman.h>

#include "mmap.h"
#include "sort.h"
#include "parse.h"
#include "mesg.h"
#include "index.h"
#include "index_local.h"
#include "index_spool.h"
#include "index_imap.h"
#include "mailbox.h"

#define d(x) x

static pthread_mutex_t local_lock;
static pthread_mutex_t spool_lock;


void mailbox_init (void)
{
	pthread_mutex_init (&local_lock, NULL);
	pthread_mutex_init (&spool_lock, NULL);
	
	pthread_mutex_lock (&local_lock);
	index_local_init ();
	pthread_mutex_unlock (&local_lock);
	
	pthread_mutex_lock (&spool_lock);
	index_spool_init ();
	pthread_mutex_unlock (&spool_lock);
	
	index_imap_init ();
}

void mailbox_commit (void)
{
	pthread_mutex_lock (&local_lock);
	index_local_commit ();
	pthread_mutex_unlock (&local_lock);
	
	pthread_mutex_lock (&spool_lock);
	index_spool_commit ();
	pthread_mutex_unlock (&spool_lock);
	
	index_imap_commit ();
}

void mailbox_garb_collect (const Mailbox_t *mailbox)
{
	GList *index_list = mailbox_get_index_list (mailbox);
	g_list_foreach (index_list, index_free, GINT_TO_POINTER (TRUE));
}

void mailbox_delete (const Mailbox_t *mailbox)
{
	gchar *mailbox_file = NULL;
	gchar *index_file = NULL;
	
	g_return_if_fail (mailbox);
	
	switch (mailbox->type) {
	case MAILBOX_LOCAL:
		pthread_mutex_lock (&local_lock);
		mailbox_file = index_local_get_location (mailbox);
		index_file = index_local_get_idx_location (mailbox);
		remove (mailbox_file);
		remove (index_file);
		pthread_mutex_unlock (&local_lock);
		break;
	case MAILBOX_SPOOL:
		/* we don't do anything here */
		break;
	case MAILBOX_IMAP:
		imap_mailbox_delete (mailbox->imap->socket, mailbox->imap->namespace, mailbox->name);
		break;
	default:
		g_assert_not_reached ();
	}
	
	if (mailbox_file)
		g_free (mailbox_file);
}

Mailbox_t *mailbox_create (const gchar *mailbox, const MAILBOX_TYPE type)
{
	Mailbox_t *mbox = NULL;
	gchar *mailbox_file = NULL;
	gint fd;
	
	g_return_val_if_fail (mailbox, NULL);
	
	mbox = g_malloc0 (sizeof (Mailbox_t));
	mbox->name = g_strdup (mailbox);
	mbox->type = type;
	
	switch (type) {
	case MAILBOX_LOCAL:
		pthread_mutex_lock (&local_lock);
		mailbox_file = index_local_get_location (mbox);
		if ((fd = open (mailbox_file, O_WRONLY | O_CREAT, 0600)) == -1) {
			d(perror ("mailbox_create::open ()"));
			g_free (mbox->name);
			g_free (mbox);
			mbox = NULL;
		} else {
			close (fd);
		}
		
		pthread_mutex_unlock (&local_lock);
		break;
	case MAILBOX_SPOOL:
		break;
	case MAILBOX_IMAP:
		break;
	default:
		g_assert_not_reached ();
	}
	
	if (mailbox)
		g_free (mailbox_file);
	
	return mbox;
}

gchar *mailbox_get_location (const Mailbox_t *mailbox)
{
	gchar *location = NULL;
	
	g_return_val_if_fail (mailbox, NULL);
	
	switch (mailbox->type) {
	case MAILBOX_LOCAL:
		pthread_mutex_lock (&local_lock);
		location = index_local_get_location (mailbox);
		pthread_mutex_unlock (&local_lock);
		break;
	case MAILBOX_SPOOL:
		pthread_mutex_lock (&spool_lock);
		location = index_spool_get_location (mailbox);
		pthread_mutex_unlock (&spool_lock);
		break;
	case MAILBOX_IMAP:
		location = index_imap_get_location (mailbox);
		break;
	default:
		g_assert_not_reached ();
	}
	
	return location;
}

Index_t *mailbox_get_index (const Mailbox_t *mailbox, const guint i)
{
	Index_t *index = NULL;
	
	g_return_val_if_fail (mailbox, NULL);
	
	switch (mailbox->type) {
	case MAILBOX_LOCAL:
		pthread_mutex_lock (&local_lock);
		index = index_local_get_index (mailbox, i);
		pthread_mutex_unlock (&local_lock);
		break;
	case MAILBOX_SPOOL:
		pthread_mutex_lock (&spool_lock);
		index = index_spool_get_index (mailbox, i);
		pthread_mutex_unlock (&spool_lock);
		break;
	case MAILBOX_IMAP:
		index = index_imap_get_index (mailbox, i);
		break;
	default:
		g_assert_not_reached ();
	}
	
	return index;
}

/* do not use this function for evil purposes */
GList *mailbox_get_index_list (const Mailbox_t *mailbox)
{
	GList *index_list = NULL;
	
	g_return_val_if_fail (mailbox, NULL);
	
	switch (mailbox->type) {
	case MAILBOX_LOCAL:
		pthread_mutex_lock (&local_lock);
		index_list = index_local_load (mailbox);
		pthread_mutex_unlock (&local_lock);
		break;
	case MAILBOX_SPOOL:
		pthread_mutex_lock (&spool_lock);
		index_list = index_spool_load (mailbox);
		pthread_mutex_unlock (&spool_lock);
		break;
	case MAILBOX_IMAP:
		index_list = index_imap_load (mailbox);
		break;
	}
	
	return index_list;
}

gchar *mailbox_get_msg (const Mailbox_t *mailbox, Index_t *index,
		       gboolean *mime, gint trim_level)
{
	gchar *msg = NULL;
	
	g_return_val_if_fail (mailbox, NULL);
	g_return_val_if_fail (index, NULL);
	
	switch (mailbox->type) {
	case MAILBOX_LOCAL:
		pthread_mutex_lock (&local_lock);
		msg = index_local_get_msg (mailbox, index);
		pthread_mutex_unlock (&local_lock);
		break;
	case MAILBOX_SPOOL:
		pthread_mutex_lock (&spool_lock);
		msg = index_spool_get_msg (mailbox, index);
		pthread_mutex_unlock (&spool_lock);
		break;
	case MAILBOX_IMAP:
		msg = index_imap_get_msg (mailbox, index);
		break;
	default:
		g_assert_not_reached ();
	}
	
	switch (mailbox->type) {
	case MAILBOX_LOCAL:
	case MAILBOX_SPOOL:
		if (mime && mime_detect (msg)) {
			*mime = TRUE;
			msg = g_strdup (msg);
			mime_parse (msg);
		} else {
			if (mime)
				*mime = FALSE;
			msg = mesg_trim_header (msg, trim_level);
		}
		break;
	case MAILBOX_IMAP:
		if (mime) {
			*mime = TRUE;
			imap_parts (mailbox->imap->socket, index_get_pos (index));
		}
		msg = g_strdup (msg);
		break;
	}
	
	return msg;
}

gchar *mailbox_get_msg_mime_part (const Mailbox_t *mailbox, Index_t *index,
				 const gint part, gint *mime_part_len)
{
	gchar *mime_part = NULL;
	gint mime_len = 0;
	gchar *msg = mailbox_get_msg (mailbox, index, NULL, 2);
	
	g_return_val_if_fail (msg, NULL);
	g_return_val_if_fail (mailbox, NULL);
	g_return_val_if_fail (index, NULL);
	
	switch (mailbox->type) {
	case MAILBOX_SPOOL:
	case MAILBOX_LOCAL:
		mime_part = mime_get_part (msg, part, &mime_len, mailbox->type, NULL);
		break;
	case MAILBOX_IMAP:
		mime_part = mime_get_part ((gchar *) index_get_pos (index), part, &mime_len,
					  mailbox->type, mailbox->imap);
		break;
	default:
		g_assert_not_reached ();
	}
	*mime_part_len = mime_len;
	g_free (msg);
	
	return mime_part;
}

guint mailbox_get_size (const Mailbox_t *mailbox)
{
	guint size = 0;
	
	g_return_val_if_fail (mailbox, 0);
	
	switch (mailbox->type) {
	case MAILBOX_LOCAL:
		pthread_mutex_lock (&local_lock);
		size = index_local_get_size (mailbox);
		pthread_mutex_unlock (&local_lock);
		break;
	case MAILBOX_SPOOL:
		pthread_mutex_lock (&spool_lock);
		size = index_spool_get_size (mailbox);
		pthread_mutex_unlock (&spool_lock);
		break;
	case MAILBOX_IMAP:
		size = index_imap_get_size (mailbox);
		break;
	default:
		g_assert_not_reached ();
	}
	
	/*d(fprintf (stderr, "Total messages for %s is %d\n", mailbox->name, size));*/
	
	return size;
}

guint mailbox_get_unread (const Mailbox_t *mailbox)
{
	guint unread = 0;

	g_return_val_if_fail (mailbox, 0);

	switch (mailbox->type) {
		case MAILBOX_LOCAL:
			pthread_mutex_lock (&local_lock);
			unread = index_local_get_unread (mailbox);
			pthread_mutex_unlock (&local_lock);
			break;
		default:
			unread = 0;
	}

	return unread;
}

Index_t *mailbox_add_msg_from_file (const Mailbox_t *mailbox, gint fd)
{
	/* here we assume the stream is in standard spruce format */
	gchar *ptr, *msg;
	struct stat st;
	Index_t *index;
	
	g_return_val_if_fail (mailbox, NULL);
	
	/* here we map the entire file, plus one extra page (zero-filled)
	 * and pass it off to mailbox_add_msg ().
	 */
	if (fstat (fd, &st) == -1) {
		d(perror ("mailbox_add_msg_from_file::fstat ()"));
		return NULL;
	}
	
	ptr = g_mmap (0, st.st_size + getpagesize (), PROT_READ, MAP_PRIVATE, fd, 0);
	
	if (ptr == (gchar *) MAP_FAILED) {
		d(perror ("mailbox_add_msg_from_file::g_mmap ()"));
		return NULL;
	}
	
	msg = g_strndup (ptr, st.st_size);
	g_munmap (ptr, st.st_size + getpagesize ());
	
	index = mailbox_add_msg (mailbox, msg);
	g_free (msg);
	
	return index;
}

Index_t *mailbox_add_msg (const Mailbox_t *mailbox, const gchar *msg)
{
	/* here we assume the stream is in standard spruce format */
	Index_t *index = NULL;
	
	g_return_val_if_fail (mailbox, NULL);
	g_return_val_if_fail (msg, NULL);
	
	switch (mailbox->type) {
	case MAILBOX_LOCAL:
		pthread_mutex_lock (&local_lock);
		index = index_local_add_msg (mailbox, msg);
		pthread_mutex_unlock (&local_lock);
		break;
	case MAILBOX_SPOOL:
		pthread_mutex_lock (&spool_lock);
		index = index_spool_add_msg (mailbox, msg);
		pthread_mutex_unlock (&spool_lock);
		break;
	case MAILBOX_IMAP:
		/*index = index_imap_add_msg (mailbox, msg);*/
		break;
	default:
		g_assert_not_reached ();
	}
	
	return index;
}

gboolean mailbox_del_msgs (const Mailbox_t *mailbox, GList *unsorted_list)
{
	gboolean result = FALSE;
	
	g_return_val_if_fail (mailbox, FALSE);
	g_return_val_if_fail (unsorted_list, FALSE);
	
	switch (mailbox->type) {
	case MAILBOX_LOCAL:
		pthread_mutex_lock (&local_lock);
		result = index_local_del_msgs (mailbox, unsorted_list);
		pthread_mutex_unlock (&local_lock);
		break;
	case MAILBOX_SPOOL:
		pthread_mutex_lock (&spool_lock);
		result = index_spool_del_msgs (mailbox, unsorted_list);
		pthread_mutex_unlock (&spool_lock);
		break;
	case MAILBOX_IMAP:
		result = index_imap_del_msgs (mailbox, unsorted_list);
		break;
	default:
		g_assert_not_reached ();
	}
	
	return result;
}

gboolean mailbox_copy_msgs (const Mailbox_t *mbox_dst, const Mailbox_t *mbox_src,
			   GList *unsorted_list)
{
	GList *sorted_list;
	
	g_return_val_if_fail (unsorted_list, FALSE);
	sorted_list = g_list_sort (unsorted_list, gint_compare);
	
	while (sorted_list) {
		Index_t *index;
		gchar *raw, *msg;
		
		index = mailbox_get_index (mbox_src, GPOINTER_TO_INT (sorted_list->data));
		g_return_val_if_fail (index, FALSE);
		
		raw = mailbox_get_msg (mbox_src, index, NULL, 2); /* get raw message */
		msg = g_strconcat (raw, "\n.\r\n", NULL);
		
		mailbox_add_msg (mbox_dst, msg);
		
		sorted_list = g_list_next (sorted_list);
		
		g_free (raw);
		g_free (msg);
	}
	
	return TRUE;
}

gboolean mailbox_move_msgs (const Mailbox_t *mbox_dst, const Mailbox_t *mbox_src,
			   GList *unsorted_list)
{
	if (!mailbox_copy_msgs (mbox_dst, mbox_src, unsorted_list)) {
		d(fprintf (stderr, "mailbox_move_msgs::failed to copy!\n"));
		return FALSE;
	} else {
		unsorted_list = g_list_first (unsorted_list);
		
		if (!mailbox_del_msgs (mbox_src, unsorted_list)) {
			d(fprintf (stderr, "mailbox_move_msgs::failed to delete!\n"));
			return FALSE;
		}
	}
	return TRUE;
}

static void mailbox_set_status_elem (gpointer key, gpointer user_data)
{
	Index_t *index = (Index_t *) key;
	gboolean status = GPOINTER_TO_INT (user_data);
	
	g_return_if_fail (index);
	index_set_seen (index, status);
}

void mailbox_set_status (const Mailbox_t *mbox, const gboolean status)
{
	GList *index_list = mailbox_get_index_list (mbox);
	g_list_foreach (index_list, mailbox_set_status_elem, GINT_TO_POINTER (status));
}
