/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include "parse.h"

GList *g_list_strtok (gchar *string, gchar *delimeters)
{
	/* Return a list of tokens based on a string and a list of delimeters;
	   tokens shall not include any delimeter characters; unlike many
	   strtok() implementations, this function does not modify the string */

	gchar *token_start, *token_end;
	gchar *delimeter, *token;
	GList *result = NULL;

	/* Return NULL if we have no string or no delimeters */
	if ((*string == '\0') || (*delimeters == '\0'))
		return NULL;

	token_start = string;

	while (*token_start) {
		/* Start after any delimeters */
		delimeter = delimeters;
		while (*delimeter && *token_start) {
			if (*delimeter == *token_start) {
				delimeter = delimeters;
				token_start++;
			} else {
				delimeter++;
			}
		}

		/* Find the end of the token */
		delimeter = delimeters;
		token_end = token_start;
		while (*token_end) {
			while (*delimeter && (*delimeter != *token_end))
				delimeter++;
	
			if (*delimeter == *token_end)
				break;
	
			token_end++;
			delimeter = delimeters;
		}

		/* Do not produce any NULL tokens */
		if (token_end != token_start) {
			token = g_strndup (token_start, (token_end - 
							 token_start));
			result = g_list_append (result, token);
		}

		token_start = token_end;
	}

	return result;
}

gint find_string (const gchar *haystack, const gchar *needle)
{           
	/* determines if one string exists inside another string.
	 * returns the location of the string if found, -1 if it is not found. */
	gchar *retval;

	if (!haystack || !needle || !*haystack || !*needle)
		return -1;

	if ( (retval = strstr(haystack, needle)) == (gchar*)NULL)
		return -1;

	return (retval - haystack);
}

gchar *nstrstr (gchar *haystack, gchar *needle, gboolean recycle)
{
	/* An implementation of Boyer Moore's string search algorythm */
	static guchar lookuptable[256];
	guint h_len, n_len, n, i;
	guchar *he_ptr, *ne_ptr;
	guchar *hc_ptr, *nc_ptr;
	guchar *h_ptr, *the_ptr;
	
	if (!haystack || !needle)
		return NULL;
	
	if ((h_len = strlen (haystack)) < (n_len = strlen (needle)))
		return NULL;
	
	/* instant match if the pattern buffer is 0-length */
	if (n_len == 0)
		return haystack;
	
	/* set a pointer at the end of each string */
	ne_ptr = needle + n_len - 1;   /* point to char before '\0' */
	he_ptr = haystack + h_len - 1; /* point to char before '\0' */
	
	if (!recycle) {
		/* if we don't want to recycle the previous lookuptable...
		 * precalculate/set look-up table */
		memset(lookuptable, n_len, 256); /* setup defaults */
		for (nc_ptr = needle; nc_ptr <= ne_ptr; nc_ptr++)
			*(lookuptable + *nc_ptr) = ne_ptr - nc_ptr;
	}
	
	h_ptr = haystack;
	while (h_len >= n_len) {
		the_ptr = h_ptr + n_len - 1;  /* set the temp haystack end pointer */
		hc_ptr = h_ptr + n_len - 1;   /* set the haystack compare pointer */
		nc_ptr = ne_ptr;              /* set the needle compare pointer */
		
		/* work our way backwards till they don't match */
		for (i = 0; nc_ptr > (guchar*)needle; nc_ptr--, hc_ptr--, i++)
			if (*nc_ptr != *hc_ptr)
				break;
		
		if (*nc_ptr != *hc_ptr) {
			n = *(lookuptable + *hc_ptr);
			if (n == 0)
				n = n_len;
			if (n == n_len && i)
				if (*ne_ptr == *needle)
					n--;
			h_ptr += n;
			h_len -= n;
		} else
			return h_ptr;
	}
	
	return NULL;
}

gchar *strstrcase (gchar *haystack, gchar *needle)
{
	/* find the needle in the haystack neglecting case */
	gchar *ptr;
	guint len;

	g_return_val_if_fail (haystack != NULL, NULL);
	g_return_val_if_fail (needle != NULL, NULL);

	if ((len = strlen (needle)) == 0)
		return haystack;
	if (len > strlen (haystack))
		return NULL;

	for (ptr = haystack; *(ptr + len - 1) != '\0'; ptr++)
		if (!g_strncasecmp (ptr, needle, len))
			return ptr;

	return NULL;
}

gchar *
strstrbound (gchar *haystack, gchar *needle, gchar *end)
{
	gboolean matches = FALSE;
	gchar *ptr;
	guint nlen;
	
	nlen = strlen (needle);
	ptr = haystack;
	
	while (ptr + nlen <= end) {
		if (!strncmp (ptr, needle, nlen)) {
			matches = TRUE;
			break;
		}
		ptr++;
	}
	
	if (matches)
		return ptr;
	else
		return NULL;
}

gchar *strrev (gchar *string)
{
	/* reverse the chars in the string */
	gchar *ptr, *eptr;

	g_return_val_if_fail (string != NULL, NULL);

	for (ptr = string, eptr = string + strlen (string) - 1; ptr < eptr; ptr++, eptr--) {
		*ptr ^= *eptr;
		*eptr ^= *ptr;
		*ptr ^= *eptr;
	}

	return string;
}

gint strinsert (gchar *string, const gchar *insertstr, guint pos)
{
	/* inserts one string into another string starting at position pos. */
	guint len;

	g_return_val_if_fail (insertstr != NULL, 0);
	g_return_val_if_fail (*insertstr != '\0', 0);
	g_return_val_if_fail (string != NULL, 0);
	
	if (*string == '\0' && pos != 0) {
		return 0;
	} else {
		if (*string == '\0' && pos == 0) {
			/* string will become the same as insertstr */
			strcpy(string, insertstr);
			return 1;
		}
	}

	len = strlen(insertstr);

	memset(string + strlen(string) + 1, 0, len + 1);
	memmove(string + pos + len, string + pos, strlen(string + pos) + 1);
	memcpy(string + pos, insertstr, len);

	return 1;
}

gint strninsert (gchar *string, const gchar *insertstr, guint pos, size_t strmax)
{
	/* inserts one string into another string starting at position pos. */
	guint len;

	g_return_val_if_fail (insertstr != NULL, 0);
	g_return_val_if_fail (*insertstr != '\0', 0);
	g_return_val_if_fail (string != NULL, 0);
	
       	if (*string == '\0' && pos != 0) {
		return 0;
	} else {
		if (*string == '\0' && pos == 0) {
			/* string will become the same as insertstr */
			strncpy(string, insertstr, strmax);
			return 1;
		}
	}

	/* out-of-bounds condition */
	if (pos >= strmax || pos > strlen(string))
		return 0;

	/* We are now left with the condition that we have a string and we have
	 * an insert string (non-NULL) and the insertion point is within bounds */

	len = strlen(insertstr);

	memset(string + strlen(string) + 1, 0, len + 1);
	memmove(string + pos + len, string + pos, strlen(string + pos) + 1);
	memcpy(string + pos, insertstr, len);

	return 1;
}

gint strcut (gchar *string, guint index, guint numchars)
{
	/* cuts a specific number of chars from a string starting at the position index */
	gchar *src, *dst;

	/* we can't cut from a NULL/empty string */
	g_return_val_if_fail (string != NULL, 0);
	g_return_val_if_fail (*string != '\0', 0);

	/* out-of-bounds condition */
	if (index >= strlen(string) || index + numchars > strlen(string))
		return 0;

	dst = string + index;
	src = dst + numchars;

	memmove(dst, src, strlen(src) + 1);

	return 1;
}

gboolean strrep (gchar *string, const gchar *remove, const gchar *replace)
{
	/* replaces the first occurence of remove with replace in string.
	 * Returns TRUE on success and FALSE on failure (meaning that
	 * `remove` could not be found within the string) */
	gint index;
   
	index = find_string(string, remove);
	if (index == -1)
		return FALSE;
	strcut(string, index, strlen(remove));
	strinsert(string, replace, index);

	return TRUE;
}

gchar *trim_leadspc (gchar *string)
{
	/* delete the leading white spaces from the string */
	gchar *strt;
	gchar *ptr1, *ptr2;

	if (string == (gchar*)NULL || *string == '\0')
		return string;
      
	/* find the first non-space and point to it with 'strt' */
	for (strt = string; *strt != '\0'; strt++)
		if (*strt != ' ' && *strt != '\t')
			break;
      
	/* copy the string beginning at 'strt' into 'string' */
	for (ptr1 = string, ptr2 = strt; *ptr2 != '\0'; ptr1++, ptr2++)
		*ptr1 = *ptr2;
	*ptr1 = '\0';

	return string;
}

gchar *trim_tailspc (gchar *string)
{
	/* delete the trailing white spaces from the string */
	gchar *end;

	if (string == (gchar*)NULL || *string == '\0')
		return string;

	/* find the end of the string... */
	for (end = string; *end != '\0'; end++);
	/* 'end' should now be pointing to the end of the string */

	end--;

	/* move backwards until we hit the first non-space char */
	for ( ; end != string; end--)
		if (*end != ' ' && *end != '\t' && *end != '\n')
			break;

	/* 'end' should now be pointing to the last char in the string */

	end++;        /* move forward one space */
	*end = '\0';  /* end the string */

	return string;
}

gchar *trim_whtspc (gchar *string)
{
	/* delete the leading/trailing white space from the string */

	if (string == (gchar*)NULL || *string == '\0')
		return string;

	trim_tailspc(string);
	trim_leadspc(string);

	return string;
}

gchar *unquote (gchar *string)
{
	/* if the string is quoted, unquote it */

	if (!string)
		return NULL;

	if (*string == '"' && *(string + strlen (string) - 1) == '"') {
		*(string + strlen (string) - 1) = '\0';
		if (*string)
			memmove (string, string+1, strlen (string));
	}

	return string;
}

gchar *strip (gchar *string, gchar c)
{
	/* strip all occurances of c from the string */
	gchar *src, *dst;

	if (string == (gchar*)NULL)
		return string;

	for (src = dst = string; *src; src++)
		if (*src != c)
			*dst++ = *src;
	*dst = '\0';

	return string;
}

gchar *parse_addr_list (gchar *list)
{
	/* change all commas (,) between addresses to semi-colons (;) */
	gchar *ptr;
	guint quotes;

	if (!list || !*list)
		return list;

	for (ptr = list, quotes = 0; *ptr; ptr++) {
		if (*ptr == '"')
			quotes++;
		else if (*ptr == ',' && !(quotes & 0x1))  /* if we're looking at a comma and the # */
			*ptr = ';';                            /* of quotes isn't odd, it's a separater */
	}

	return list;
}

gchar *unparse_addr_list (gchar *list)
{
	/* change all semi-colons (;) between addresses to commas (,) */
	gchar *ptr;
	guint quotes;

	if (!list || !*list)
		return list;

	for (ptr = list, quotes = 0; *ptr; ptr++) {
		if (*ptr == '"')
			quotes++;
		else if (*ptr == ';' && !(quotes & 0x1))  /* if we're looking at a sem-colon and the # */
			*ptr = ',';                            /* of quotes isn't odd, it's a separater */
	}

	return list;
}

guint get_num_fields (gchar *string, gchar delim)
{
	/* returns the number of fields in a string
	 * assumes that the delimeter does not appear
	 * at the beginning nor the end of the string. */
	guint count = 1;
	gchar *ptr;

	if (string == (gchar*)NULL || *string == '\0')
		return 0;

	trim_whtspc(string);

	if (*string == '\0')
		return 0;

	for (ptr = string; *ptr != '\0'; ptr++)
		if (*ptr == delim)
			count++;

	return count;
}

gchar *get_field (gchar *string, guint field, gchar delim)
{
	gchar *buffer = NULL;
	gchar *start, *ptr1;
	guint cfield = 1;

	if (string == NULL || *string == '\0')
		return NULL;
	
	if (field <= 0)
		return NULL;
	
	/* set the starting position to the beginng of the string */
	start = string;
	
	/* Already on the field we want...? */
	if (cfield != field) {
		for ( ; *start != '\0'; start++)   /* keep advancing chars... */
			if (*start == delim) {
				/* when we hit a delimiter... */
				cfield++;  /* increment the current field */
				if (cfield >= field)
					break; /* at the right field */
			}
	}
	
	if (field != 1)
		start++;  /* advace to the char after the delimiting char */
	
	for (ptr1 = start; *ptr1 && *ptr1 != delim; ptr1++);
	
	buffer = g_strndup (start, (gint)(ptr1 - start));
	
	return buffer;
}

/* rfc1522 says that if you want 8bit chars in the headers
 * you have to encode it in a special way:
 * =?charset?encoding?text?=
 * the text goes into ?text? with the 8bit characters escaped
 * to =<hexcode> , for example =F6 would be .
 */
/* this whole function is ugly =) /kristoffer */
  
gint parse_8bit (gchar *text)
{
	gchar buffer[128], *decoded = NULL;
	gint len, wordlen = 0, index, pos, i, q = 0;
	gint equal = 0;

	if (text == (gchar*)NULL || *text == '\0')
		return 0;

	memset(buffer, 0, sizeof(buffer));
	len = strlen(text);
	i = 0;
	while (i < len) {
		if (equal) {
			if (text[i] == '?') {
				index = i - 1;
				i++;
				pos = 0;
				wordlen = 0;
				q = 0;
				while (i < len) {
					buffer[pos] = text[i];
					if (buffer[pos] == '?')
						q++;
					pos++;
					wordlen++;
					i++;
					if (q == 2)
						break;
				}
				while (i < len && (text[i] != '?' || text[i+1] != '=')) {
					buffer[pos] = text[i];
					pos++;
					wordlen++;
					i++;
				}
				buffer[pos] = '\0';
				pos = 0;
				/* ignore the charset and continue to the encoding */
				while (buffer[pos] != '\0' && buffer[pos] != '?')
					pos++;
				
				pos++;

				/* encoding types */
				if (buffer[pos+1] == '?') {
					switch (buffer[pos]) {
					case 'B':
					case 'b':
						/* base64 */
						strcut(buffer, 0, pos + 2);
						if (!(pos = strlen(buffer)))
							return 0;   /* something went wrong */
						decoded = decode_base64(buffer, &pos);
						break;
					case 'Q':
					case 'q':
						/* quoted printable */
						strcut(buffer, 0, pos + 2);
						if (!(pos = strlen(buffer)))
							return 0;   /* something went wrong */
						decoded = decode_quoted_printable(buffer, &pos);
						for (pos = 0; decoded[pos] != '\0'; pos++)
							if (decoded[pos] == '_')
								decoded[pos] = ' ';
						break;
					default:
						/* unknown */
						return 0;
					}
				}
				
				strcut(text, index, wordlen + 4);
				strinsert(text, decoded, index);

				/* ok, i must be changed because it points at the end of the encoded data
				 * which is has now been replaced with the decoded data */
				i = index + strlen(decoded) - 1;
				len = strlen(text);
				g_free(decoded);
			}
			equal = 0;
		} else if (text[i] == '=') {
			equal = 1;
		}
		i++;
	}
	
	return 1;
}

/* if there are 8bit chars in the string, encode it */
gchar *make_8bit (guchar *text)
{
	guint len, size;
	gint i = 0, y = 0, pos = 0, wordlen;
	gint eightbit = 0;
	guchar *buffer = NULL;
	gchar temp[6];
	guchar *startword, *endword;
	guchar *p;
	
	if (text == NULL || *text == '\0')
		return NULL;
	
	size = len = strlen(text);
	size++;
	buffer = g_malloc0 (size);
	p = text;

	while (i < len) {
		startword = p;
		while (*p != ' ' && *p != '\0' && i < len) {
			p++;
			i++;
		}
		endword = p;
		wordlen = (endword - startword);
		i = i - wordlen;
      
		y = 0;
		while (y < wordlen) {
			if (text[i] > 127)
				eightbit++;
			i++;
			y++;
		}
		i = i - wordlen;
		
		if (eightbit > 0) {
			size = size + eightbit * 3 + 17;
			buffer = g_realloc (buffer, size);
			buffer[pos] = '\0';
			strcat(buffer, "=?iso-8859-1?Q?");
			pos += 15;
			y = 0;
			
			while (y < wordlen) {
				if (text[i] > 127 || text[i] == '=') {
					sprintf(temp, "%X", text[i]);
					buffer[pos] = '=';
					buffer[pos+1] = temp[0];
					buffer[pos+2] = temp[1];
					pos += 3;
				} else {
					if (text[i] == ' ') {
						buffer[pos] = '_';
						pos++;
					} else {
						buffer[pos] = text[i];
						pos++;
					}
				}

				i++;
				y++;
			}
			buffer[pos] = '?';
			buffer[pos+1] = '=';
			pos += 2;
		} else {
			y = 0;
			while (y < wordlen) {
				buffer[pos] = text[i];
				pos++;
				i++;
				y++;
			}
		}
      
		if (i >= len || *p == '\0')
			break;
         
		eightbit = 0;
		buffer[pos] = ' ';
		pos++;
		i++;
		p++;
	}

	buffer[pos] = '\0';

	return buffer;
}

guint mesg_get_longest_line_len (gchar *mesg, guint tab_equiv)
{
	gchar *curpos;
	guint longest_line_len, current_line_len;

	g_return_val_if_fail (mesg, 0);

	longest_line_len = 0;
	current_line_len = 0;

	for (curpos = mesg; *curpos; curpos++) {
		if (*curpos == '\n') {
			if (current_line_len > longest_line_len)
				longest_line_len = current_line_len;
			current_line_len = 0;
		} else {
			if (*curpos == '\t') {
				current_line_len += (tab_equiv - 
				    (current_line_len % tab_equiv));
			} else
				current_line_len++;
		}
	}

	/* Catch lines which are not followed by a newline */
	if (current_line_len > longest_line_len)
		longest_line_len = current_line_len;

	return longest_line_len;
}

guint mesg_get_tab_count (gchar *mesg)
{
	gchar *curpos;
	guint tab_count;

	g_return_val_if_fail (mesg, 0);

	tab_count = 0;

	for (curpos = mesg; *curpos; curpos++) {
		if (*curpos == '\t')
			tab_count++;
	}

	return tab_count;
}

gchar *soft_word_wrap (gchar *text, guint len, guint tab_equiv)
{
	/* Perform word wrap, not breaking up words, treating tabs
	as tab sets every 'tab_equiv' columns and wrapping every 'len' chars */
	
	gchar *curpos, *lastspc, *lastbrk, *laststart;
	guint current_line_len;

	/* If we've been told to wrap at 0 chars, do no wrapping */
	if (len == 0)
		return text;

	if (tab_equiv < 1)
		tab_equiv = 1;

	current_line_len = 0;
	curpos = lastspc = lastbrk = laststart = text;

	while (*curpos) {
		if (*curpos == ' ') {
			lastspc = curpos;
			current_line_len++;
		} else if (*curpos == '\t') {
			lastspc = curpos;
			current_line_len += tab_equiv - (current_line_len % tab_equiv);
		} else if (*curpos == '\n') {
			lastbrk = curpos;
			lastspc = lastbrk;
			current_line_len = 0;
		} else {
			current_line_len++;
		}
		
		/* laststart prevents us from attempting to wrap the same
		   line twice; curpos is designed to move both forward and
		   backward over the string, so as to perform the best wrap */
		if ((current_line_len >= len) && (laststart != lastspc)) {
			*lastspc = '\n';
			lastbrk = lastspc;
			curpos = lastbrk;
			laststart = lastbrk;
			current_line_len = 0;
		}

		curpos++;
	}

	return text;
}

gchar *word_wrap (gchar *text, guint len, guint tab_equiv, guint tab_convert,
		  guint det_pgp_clearsign)
{
	/* REMEMBER to free the memory returned by this function! */
	gchar *modtext, *curpos, *modpos;
	guint bufferlen, current_line_length, tabs_to_insert, i;

	/* Check for pgp clearsign detection flag */
	if (det_pgp_clearsign) {
		if (strstr(text, "-----BEGIN PGP SIGNED MESSAGE-----"))
			return g_strdup (text);
	}

	/* Perform tab conversion */
	if (tab_convert) {
		bufferlen = (strlen (text) + (mesg_get_tab_count (text) 
					      * tab_equiv) + 1);
		modtext = g_malloc0 (bufferlen);

		current_line_length = 0;
		curpos = text;
		modpos = modtext;

		while (*curpos) {
			if (*curpos == '\t') {
				tabs_to_insert = tab_equiv - 
				    (current_line_length % tab_equiv);
				for (i = 0; i < tabs_to_insert; i++) {
					*modpos = ' ';
					modpos++;
				}
				current_line_length += tabs_to_insert;
			} else if (*curpos == '\n') {
				*modpos = *curpos;
				modpos++;
				current_line_length = 0;
			} else {
				*modpos = *curpos;
				modpos++;
				current_line_length++;
			}

			curpos++;
		}
	} else {
		modtext = g_strdup (text);
	}

	return soft_word_wrap(modtext, len, tab_equiv);
}

gchar *quote_body (gchar *body, gchar *rcpt, gchar *date)
{
	gchar *new_body;
	gchar *src, *dst, *new_date, *name;
	gint lines = 0;
	gint length;

	/* Count the number of lines in the body. */
	for (src = body; *src; src++)
		if (*src == '\n')
			lines++;

	new_date = (gchar *) g_malloc0(strlen(date) + 1);
	dst = new_date;
	/* keep copying over until we hit the first ':' in the time stamp... */
	for (src = date; *src; src++) {
		if (*src == ':')
			break;
		*dst = *src;
		dst++;
	}
	/* ...now back-track... */
	for ( ; dst != new_date; dst--)
		if (*dst == ' ')
			break;
	*dst = '\0';

	/* if avail, get the user's name or else just use his email address */
	name = (gchar *) g_malloc0(strlen(rcpt) + 1);
	dst = name;
	for (src = rcpt; *src; src++) {
		if (*src == '<' && src != rcpt)
			break;
		*dst = *src;
		dst++;
	}
	*dst = ' ';
	for ( ; dst != name; dst--)
		if (*dst == ' ')
			*dst = '\0';
		else
			break;

	/* unquote the sender's name, makes it look prettier ;-) */
	unquote(name);

	/* Allocate space for the new body. */
	length = (2 * (lines + 1)) + strlen(body) + strlen(name) + strlen(new_date) + 16;
	new_body = (gchar *)g_malloc0(length);
	if (new_body == 0) {
		g_free(name);
		g_free(new_date);
		return NULL;
	}

	/* On $new_date, $recpt wrote: */
	strcpy (new_body, "\n\nOn ");
	strcat (new_body, new_date);
	strcat (new_body, ", ");
	strcat (new_body, name);
	strcat (new_body, " wrote:\n"); 

	/* just in case the name is 8-bit quoted printable... */
	parse_8bit(new_body);

	/* Reformat the text with leading "> ". */
	dst = new_body + strlen(new_body);
	*dst = '>';
	dst++;
	*dst = ' ';
	dst++;
	for (src = body; *src; src++) {
		*dst = *src;
		dst++;
		if (*src == '\n') {
			*dst = '>';
			dst++;
			*dst = ' ';
			dst++;
		}
	}
	/* *(dst - 2) = '\0'; */

	g_free(name);
	g_free(new_date);

	return new_body;
}

gint getopts_short (gint argc, gchar *argv[], gchar *optstring)
{
	/* The getopts_short() function parses the command line arguments.
	 * Its arguments argc and argv are the argument count and
	 * array as passed to the main() function on program invocation.
	 * An element of argv that starts with `-' (and is not exactly
	 * "-" or "--") is an option element. The characters of this
	 * element (aside from the initial `-') are option characters.
	 * The third argument, optstring, holds the valid list of
	 * option characters. If a character in an option element is
	 * found not to be included in optstring, then 0 is returned.
	 * If getopts_short() is called repeatedly, it returns
	 * successively each of the option characters from each of
	 * the option elements. If there are no more option characters,
	 * getopts_short() returns EOF (-1). */

	gint n, i;
	gchar *s_opt, *arg_ptr;
	static guint nextchar = 0;
	static guint shortindex = 0;

	for (n = shortindex + 1; n < argc; n++) {
		if (*argv[n] == '-' && *(argv[n]+1) != '-') {
			/* we have ourselves a short option */
			arg_ptr = argv[n];
			arg_ptr++;  /* advance to the first char after the '-' */

			/* advance to where we left off */
			for (i = 0; i < nextchar; arg_ptr++, i++);

			for (s_opt = optstring; *s_opt != '\0'; s_opt++) {
				if (*s_opt == *arg_ptr) {
					if ( *(s_opt+1) == '\0' && *(arg_ptr+1) == '\0') {
						shortindex = n;  /* up the index... */
						nextchar = 0;    /* and reset nextchar */
					} else {
						nextchar++;      /* else remember the next char */
					}
					
					return (gint) *s_opt;
				}
			}
			
			if (*s_opt == '\0' && *arg_ptr != '\0') {
				/* Huston, we have a problem... This is an invalid arg */
				fprintf (stderr, "%s: invalid option [-%c]\n", argv[0], *arg_ptr);
				fflush(stderr);
				nextchar++;
				return 0;
			}
		}
		nextchar = 0;
		shortindex++;
	}

	return -1;
}

gint getopts_long (gint argc, gchar *argv[], gchar *optstring,
                   longopts_t *longopts, gint *longindex, gchar **param)
{
	/* The getopts_long() function works like getopts_short()
	 * except that it also accepts long options, started out 
	 * by two dashes ("--"). A long option may take a
	 * parameter. This must be of the form "--arg=param" or 
	 * "--arg param". If the long option takes a parameter,
	 * param will point to the beginning of the parameter
	 * inside of argv. Longindex holds the current position
	 * in argv. You may use this to skip around inside of argv
	 * or to skip the processing of any short options by
	 * setting it to a non-zero value. In most cases, this should
	 * be set to 0 on the first call for this function so that
	 * no options are missed. */

	gint n;
	gchar l_name[256], *l_ptr, *ptr;
	longopts_t *l_opt;
	gboolean used_equal = FALSE;

	*param = NULL;

	/* get short options first (if we haven't started to get the long options yet) */
	if (*longindex == 0)
		if ( (n = getopts_short(argc, argv, optstring)) != -1)
			return n;

	/* okay, we made it here...so getopt() must have gotten all the short opts */
	for (n = *longindex + 1; n < argc; n++) {
		if (!strncmp("--", argv[n], 2)) {
			/* looks like we found a long option */
			memset(l_name, 0, sizeof(l_name));
			l_ptr = argv[n];
			l_ptr += 2; /* advance to just beyond the "--" */
			strncpy(l_name, l_ptr, sizeof(l_name)-1);
         
			used_equal = FALSE;
			/* search and end the string at an '=' assuming there is one */
			for (ptr = l_name; *ptr != '\0'; ptr++) {
				if (*ptr == '=') {
					*ptr = '\0';
					used_equal = TRUE;
				}
			}

			for (l_opt = longopts; l_opt->name != NULL; l_opt++) {
				if (!strcmp(l_opt->name, l_name)) {
					/* we found a match */
					*longindex = n;

					switch (l_opt->has_arg) {
					case no_arg:
						return l_opt->val;          /* return val */
						break;
					case required_arg:
						if (used_equal) {
							for (ptr = argv[n]; *ptr != '='; ptr++);
							ptr++;
							*param = ptr; /* set the param */
						} else {
							*longindex = n + 1;
							if (n+1 < argc) {
								/* the param is the next element */
								*param = argv[n+1];
							} else {
								fprintf (stderr, "Warning: %s: incomplete option.\n",
									 l_name);
								fflush(stderr);
								*param = NULL;    /* nothing to point to */
							}
						}
						
						return l_opt->val;
						break;
					case optional_arg:
						if (used_equal) {
							/* yep, we had an arg */
							for (ptr = argv[n]; *ptr != '='; ptr++);
							ptr++;
							*param = ptr; /* param is after the '=' */
						} else {
							/* maybe we have an arg, or maybe not */
							/* if this fails, we can't have an arg */
							if (n+1 < argc) {
								/* if this fails, no arg */
								if (argv[n+1][0] != '-') {
									/* ladies and gents, we have an arg */
									*longindex = n + 1;
									*param = argv[n+1];
								}
							}
						}
						
						return l_opt->val;
					}
				}
			}

			fprintf (stderr, "%s: invalid option [--%s]\n", argv[0], l_name);
			fflush(stderr);
			*longindex = n;
			return 0;
			break;
		}
	}

	return -1;
}
