/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include "pop3io.h"
#include "uid.h"

#define d(x) x

extern GtkWidget *frmStatus;

extern guint timeout;

static gboolean pop_auth (gint sock, gchar *login, gchar *password, SpruceException *ex);
static gboolean pop_apop (gint sock, gchar *login, gchar *password, gchar *challenge, SpruceException *ex);
static gboolean pop_user (gint sock, gchar *login, SpruceException *ex);
static gboolean pop_pass (gint sock, gchar *passwd, SpruceException *ex);
static gint pop_stat (gint sock, SpruceException *ex);
static gint *pop_stat_trim (gint sock, gint *num, SpruceException *ex);
static gint pop_list (gint sock, SpruceException *ex);
static gboolean pop_uidl (gint sock, guint id, gchar *uid, SpruceException *ex);
static gboolean pop_retr (gint sock, guint id, Mailbox_t *finbox, gboolean del_yn, SpruceException *ex);
static gboolean pop_dele (gint sock, guint id, SpruceException *ex);
static gboolean pop_rset (gint sock, SpruceException *ex);
static gboolean pop_quit (gint sock, SpruceException *ex);


void emesg_destroy (Emesg *msg)
{
	/* destroy the mem space this email message is taking up */
	g_free (msg->header);
	msg->header = NULL;
	g_free (msg->body);
	msg->body = NULL;
}

static gboolean
pop_auth (gint sock, gchar *login, gchar *password, SpruceException *ex)
{
	gchar *cmdbuf;
	gchar buffer[512];
	gchar *retstr;
	gint pos;
	
	cmdbuf = g_strdup ("AUTH CRAM-MD5\r\n");
	
	if (SSLSend (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
				       _("CRAM-MD5 AUTH request timed out: %s"),
				       g_strerror (errno));
		return FALSE;
	}
	g_free (cmdbuf);
	
	memset (buffer, 0, sizeof (buffer));
	if ((recvline (sock, buffer, sizeof (buffer)-1) < 0) || !strncmp (buffer, "-ERR", 4)) {
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
				       _("CRAM-MD5 AUTH response error: %s"),
				       *buffer ? buffer : g_strerror (errno));
		return FALSE;
	}
	
	pos = find_string (buffer, "\r\n");
	if (pos != -1)
		buffer[pos] = '\0';
	retstr = cram_md5 (login, password, buffer);
	
	if (retstr[strlen (retstr)-1] == '\n')
		retstr[strlen (retstr)-1] = '\0';
	
	cmdbuf = g_strdup_printf ("%s\r\n", retstr); 
	if (SSLSend (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (retstr);
		g_free (cmdbuf);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
				       _("CRAM-MD5 authentication request timed out: %s"),
				       g_strerror (errno));
		return FALSE;
	}
	g_free (cmdbuf);
	g_free (retstr);
	
	memset (buffer, 0, sizeof (buffer));
	if ((recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "+OK", 3)) {
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
				       _("Failed to authenticate using CRAM-MD5: %s"),
				       *buffer ? buffer : g_strerror (errno));
		return FALSE;
	}
	
	return TRUE;
}

static gboolean
pop_apop (gint sock, gchar *login, gchar *password, gchar *challenge, SpruceException *ex)
{
	gchar *cmdbuf;
	gchar buffer[512];
	md5_state_t md5_state;
	md5_byte_t digest[16];
	gchar asciidigest[33];
	gint i;
	
	md5_init (&md5_state);
	
	i = find_string (challenge, "\r\n");
	if (i != -1)
		challenge[i] = '\0';
	
	md5_append (&md5_state, challenge, strlen (challenge));
	md5_append (&md5_state, password, strlen (password));
	md5_finish (&md5_state, digest);
	
	for (i = 0; i < 16; i++)
		sprintf (asciidigest+2*i, "%02x", digest[i]);
	
	cmdbuf = g_strdup_printf ("APOP %s %s\r\n", login, asciidigest);
	
	if (SSLSend (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
				       _("APOP authentication request timed out: %s"),
				       g_strerror (errno));
		return FALSE;
	}
	g_free (cmdbuf);
	
	memset (buffer, 0, sizeof (buffer));
	if ((recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "+OK", 3)) {
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
				       _("APOP authentication response error: %s"),
				       *buffer ? buffer : g_strerror (errno));
		return FALSE;
	}
	
	return TRUE;
}

static gboolean
pop_user (gint sock, gchar *login, SpruceException *ex)
{
	/* login to the server - step 1 (send login name) */
	gchar buffer[512];
	gchar *cmdbuf;
	
	/* lets send our user login to the server */
	cmdbuf = g_strdup_printf ("USER %s\r\n", login);
   
	if (SSLSend (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
				       _("USER request timed out: %s"),
				       g_strerror (errno));
		return FALSE;
	}
	g_free (cmdbuf);
	
	memset(buffer, 0, sizeof (buffer));
	if ((recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "+OK", 3)) {
		/* I'm guessing the pop3 server didn't like our USER command... */
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
				       _("APOP authentication response error: %s"),
				       *buffer ? buffer : g_strerror (errno));
		return FALSE;
	}
	
	return TRUE;
}

static gboolean
pop_pass (gint sock, gchar *passwd, SpruceException *ex)
{
	/* login to the server - step 2 (send passwd) */
	gchar buffer[512];
	gchar *cmdbuf;
	
	/* lets send our user passwd to the server */
	cmdbuf = g_strdup_printf ("PASS %s\r\n", passwd);

	if (SSLSend (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
				       _("PASS request timed out: %s"),
				       g_strerror (errno));
	}
	g_free (cmdbuf);
	
	memset (buffer, 0, sizeof (buffer));
	if ((recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "+OK", 3)) {
		/* I'm guessing the pop3 server didn't like our PASS command... */
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
				       _("PASS response error: %s"),
				       *buffer ? buffer : g_strerror (errno));
		return FALSE;
	}
	
	return TRUE;
}

static gint
pop_stat (gint sock, SpruceException *ex)
{
	/* returns the number of messages on the server */
	gchar buffer[512];
	gchar *cmdbuf;
	gint num;
	
	/* lets STAT the pop3 server... */
	cmdbuf = g_strdup ("STAT\r\n");
	
	if (SSLSend (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("STAT request timed out: %s"),
				       g_strerror (errno));
		return -1;
	}
	g_free (cmdbuf);
	
	memset (buffer, 0, sizeof (buffer));
	if ((recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "+OK", 3)) {
		/* something went wrong... */
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("STAT response error: %s"),
				       *buffer ? buffer : g_strerror (errno));
		return -1;
	}
	
	/* get the number of messages on the server */
	sscanf (buffer, "+OK %d messages\r\n", &num);
	
	return num;
}

static gint *
pop_stat_trim (gint sock, gint *num, SpruceException *ex)
{
	gchar buffer[512];
	gchar *cmdbuf;
	gchar uid[100];
	gint id, count = 0;
	gint msglist[*num], *msgptr = NULL;
	
	/* send UIDL to server to get full list of UIDs */
	cmdbuf = g_strdup ("UIDL\r\n");
	if (SSLSend (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("UIDL request timed out: %s"),
				       g_strerror (errno));
		return NULL;
	}
	g_free (cmdbuf);
	
	memset (buffer, 0, sizeof (buffer));
	if (recvline (sock, buffer, sizeof (buffer)-1) < 0) {
		d(fprintf (stderr, _("Unable to read UIDL response\n")));
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("UIDL response error: %s"),
				       *buffer ? buffer : g_strerror (errno));
		return NULL;
	}
	
	/* first response line should just be +OK */
	if (find_string (buffer, "+OK") < 0) {
		g_strstrip (buffer);
		d(fprintf (stderr, _("Unrecognized UIDL respose: \"%s\"\n"), buffer));
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("Unrecognized UIDL respose: %s"),
				       buffer);
		return NULL;
	}
	
	/* now parse UIDlist, format is "# uid" */
	memset (buffer, 0, sizeof (buffer));
	while ((recvline (sock, buffer, sizeof (buffer)-1) > 0) && strncmp (buffer, ".\r\n", 3)) {
		/* grab UID */
		if (sscanf (buffer, "%d %s", &id, uid) < 2) {
			g_strstrip (buffer);
			d(fprintf (stderr, _("Invalid UIDL response: \"%s\"\n"), buffer));
			continue;
		}
		
		/* bounds checking on mailbox size */
		if (id <= *num && !uid_lookup (uid)) {
			msglist[count++] = id;
		}
	}
	*num = count;
	
	/* make a list--only large enough for unread msgs */
	if (count) {
		msgptr = g_malloc0 (count * sizeof (gint));
		if (msgptr == NULL) {
			d(fprintf (stderr, "pop3io: out-of-memory\n"));
			return NULL;
		}
		memcpy (msgptr, &msglist, count * sizeof (gint));
	}
	
	return msgptr;
}

static gint
pop_list (gint sock, SpruceException *ex)
{
	/* list the messages on the server 
	 * returns the number of messages */
	gchar buffer[512];
	gchar *cmdbuf;
	gint num = 0;
	
	/* prepare the LIST command */
	cmdbuf = g_strdup ("LIST\r\n");
   
	if (SSLSend (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("LIST request timed out: %s"),
				       g_strerror (errno));
		return 0;
	}
	g_free (cmdbuf);
	
	memset (buffer, 0, sizeof (buffer));
	if ((recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "+OK", 3)) {
		/* something went wrong */
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("LIST response error: %s"),
				       *buffer ? buffer : g_strerror (errno));
		return 0;
	}
	
	while (recvline (sock, buffer, sizeof (buffer)-1) > 0) {
		num++;
	}
	
	num--;
	
	return num;
}

static gboolean
pop_uidl (gint sock, guint id, gchar *uid, SpruceException *ex)
{
	/* get the uid of the message */
	gchar buffer[512];
	gchar *cmdbuf;
	gint id2;
	gint n;
	
	/* lets tell the server which message to retrieve */
	cmdbuf = g_strdup_printf ("UIDL %d\r\n", id);
	
	if (SSLSend (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("UIDL request timed out: %s"),
				       g_strerror (errno));
		return FALSE;
	}
	g_free (cmdbuf);
	
	memset (buffer, 0, sizeof (buffer));
	if ((n = recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "+OK", 3)) {
		if (n >= 0) {
			g_strstrip (buffer);
			d(fprintf (stderr, _("UIDL expected \"+OK\" but found: \"%s\"\n"), buffer));
		}
		
		/* Maybe that message doesn't exist? Or maybe something else got fux0red */
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("UIDL response error: %s"),
				       n < 0 ? g_strerror (errno) : buffer);
		return FALSE;
	}
	
	sscanf (buffer, "+OK %d %s", &id2, uid);
	
	return TRUE;
}

static gboolean
pop_retr (gint sock, guint id, Mailbox_t *finbox, gboolean del_yn, SpruceException *ex)
{
	/* retrieve a specific email message off the server */
	gchar buffer[512];
	gchar *cmdbuf;
	gchar uid[100];
	char *template;
	Mailbox_t mailbox;
	gchar *tmbx;
	gint n;
	gint tmpfd;
	
	/* generate a somewhat random file name */
	template = g_strdup_printf ("spruce-pop-retr-%d-%d", (int) time (NULL), getpid ());
	
	mailbox.name = template;
	mailbox.type = MAILBOX_LOCAL;
	tmbx = mailbox_get_location (&mailbox);
	
	/* we don't need this anymore */
	g_free (template);
	
	/* lets find out if we already have this message */
	if (pop_uidl (sock, id, uid, ex)) {
		if (uid_lookup (uid))
			return TRUE;
	}
	
	/* lets tell the server which message to retrieve */
	cmdbuf = g_strdup_printf ("RETR %d\r\n", id);
	
	if (SSLSend (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("RETR request timed out: %s"),
				       g_strerror (errno));
	}
	g_free (cmdbuf);
	
	memset (buffer, 0, sizeof (buffer));
	if ((n = recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "+OK", 3)) {
		if (n >= 0) {
			g_strstrip (buffer);
			d(fprintf (stderr, _("RETR expected \"+OK\" but found: \"%s\"\n"), buffer));
		}
		/* Maybe that message doesn't exist? Or maybe something else got fux0red */
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("RETR response error: %s"),
				       n < 0 ? g_strerror (errno) : buffer);
		g_free (tmbx);
		return FALSE;
	}
	
	/* first we get the mail to a temp file, then run a filter check on it
	 * if it's not affected by any filters it goes to the normal mailbox. */
	tmpfd = open (tmbx, O_CREAT /*| O_SYNC*/ | O_TRUNC | O_RDWR, 0600); 
	if (tmpfd == -1) {
		g_free (tmbx);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("Failed to open temporary mailbox: %s"),
				       g_strerror (errno));
		return FALSE;
	}
	
	memset (buffer, 0, sizeof (buffer));
	n = recvline (sock, buffer, sizeof (buffer)-1);
	
	while ((n > 0) && strncmp (buffer, ".\r\n", 3)) {
		strip (buffer, '\r');  /* strip all the \r's */
		write (tmpfd, buffer, strlen (buffer));
		memset (buffer, 0, sizeof (buffer));
		n = recvline (sock, buffer, sizeof (buffer)-1);
	}
	
	write (tmpfd, ".\r\n", strlen (".\r\n"));
	
	if (!check_filters (tmpfd)) {
		lseek (tmpfd, 0, SEEK_SET);
		mailbox_add_msg_from_file (finbox, tmpfd);
	}
	close (tmpfd);
	unlink (tmbx);
	
	g_free (tmbx);
	
	/* since we finished getting this mail message
	 * lets add it to the uid list so that we know to
	 * not download it next time */
	
	if (!del_yn) {
		/* if we're deleting this mesg, no use saving it's uid */
		uid_add (uid);
	}
	
	return TRUE;
}

static gboolean
pop_dele (gint sock, guint id, SpruceException *ex)
{
	/* delete message #id off the pop3 server */
	gchar buffer[512];
	gchar *cmdbuf;
	
	/* lets tell the server which message to delete */
	cmdbuf = g_strdup_printf ("DELE %d\r\n", id);
	
	if (SSLSend (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("DELE request timed out: %s"),
				       g_strerror (errno));
	}
	g_free (cmdbuf);
	
	memset (buffer, 0, sizeof (buffer));
	if (recvline (sock, buffer, sizeof (buffer)-1) < 0 || strncmp (buffer, "+OK", 3)) {
		/* Maybe that message doesn't exist? Or maybe something else got fux0red */
		g_strstrip (buffer);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("DELE response error: %s"),
				       *buffer ? buffer : g_strerror (errno));
	}
	
	return TRUE;
}

static gboolean
pop_rset (gint sock, SpruceException *ex)
{
	/* we are going to reset the smtp server (just to be nice) */
	gchar buffer[512];
	gchar *cmdbuf;
	
	cmdbuf = g_strdup ("RSET\r\n");
	
	if (SSLSend (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("RSET request timed out: %s"),
				       g_strerror (errno));
	}
	g_free (cmdbuf);
	
	memset (buffer, 0, sizeof (buffer));
	if (recvline (sock, buffer, sizeof (buffer)-1) < 0 || strncmp (buffer, "+OK", 3)) {
		/* the server should have told us that it's been reset */
		g_strstrip (buffer);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("RSET response error: %s"),
				       *buffer ? buffer : g_strerror (errno));
	}
	
	return TRUE;
}

static gboolean
pop_quit (gint sock, SpruceException *ex)
{
	/* lets tell the server we want to quit now */
	gchar buffer[256];
	gchar *cmdbuf;
	
	cmdbuf = g_strdup ("QUIT\r\n");
	
	if (SSLSend (sock, cmdbuf, strlen (cmdbuf), 0) < 0) {
		g_free (cmdbuf);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("QUIT request timed out: %s"),
				       g_strerror (errno));
	}
	g_free (cmdbuf);
	
	memset (buffer, 0, sizeof (buffer));
	if (recvline (sock, buffer, sizeof (buffer)-1) < 0 || strncmp (buffer, "+OK", 3)) {
		/* the server should have told us that it's been reset */
		g_strstrip (buffer);
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SYSTEM,
				       _("QUIT response error: %s"),
				       *buffer ? buffer : g_strerror (errno));
	}
	
	return TRUE;
}

gint fetchmail (Server *server, gchar *login, gchar *passwd, Mailbox_t *finbox, 
		gboolean del_yn, gboolean show_dialog, conn_auth_t authtype,
		gboolean useSSL, SpruceException *ex)
{
	/* this will interface with a pop3 daemon on a foreign host
	 * and retrieve any new messages on the server.
	 * "finbox" is the file to append new messages to (your Inbox)
	 * "del_yn" specifies whether or not to delete the messages on
	 * server once they are retrieved.
	 */
	
	gint sock;
	gint num;
	guint i, fetched = 0;
	gint challenge_pos;
	gchar buffer[256];
	gfloat progress;
	gint *msglist = NULL;
	
	/* resolve the hostname */
	if (show_dialog)
		invoke_in_gui_thread (mailthread_update_status_text, 
				      _("Resolving server address..."));
	
	if (!Resolve (server)) {
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_UNAVAILABLE,
				       "Hostname (%s) failed to resolve: %s",
				       server->hostname,
				       g_strerror (errno));
		return 0;
	}
	
	/* assign and open a socket */
	sock = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
	if (sock < 0) {
		/* socket failed to open */
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_UNAVAILABLE,
				       "Socket failed to open: %s",
				       g_strerror (errno));
		return 0;
	}
	
	/* lets set up information about the server we will be connecting to... */
	server->sin.sin_family = AF_INET;
	server->sin.sin_port = htons (server->port);
	
	if (show_dialog) {
		invoke_in_gui_thread (mailthread_update_status_text, "Connecting to server...");
	}
	
	/* call the server */
	if (connect_timeo (sock, (struct sockaddr*)&server->sin, sizeof (server->sin), timeout) < 0) {
		/* we failed to connect */
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_UNAVAILABLE,
				       "Failed to connect to host (%s): %s",
				       server->hostname,
				       g_strerror (errno));
		close (sock);
		return 0;
	}
	
#ifdef WITH_SSL
	if (useSSL) {
		if (openSSLConnection (sock) < 0) {
			d(fprintf (stderr, "Failed to open SSL connection!\n"));
			spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_UNAVAILABLE,
				       "Failed to open a secure connection: %s",
				       g_strerror (errno));
			close (sock);
			return 0;
		}
	}
#endif	
	
	if ((recvline (sock, buffer, sizeof (buffer)-1) < 0) || strncmp (buffer, "+OK", 3)) {
		/* the pop3 server failed to send a response (usually version info) */
		spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_UNAVAILABLE,
				       "Server failed to acknowledge: %s",
				       g_strerror (errno));
		close (sock);
		return 0;
	}
	
	/* In case the daemon wants to give us more info than we expect...
	 * basically what we do here is wait for recv() to return an error code
	 * (EWOULDBLOCK) which would tell us that the wait timed out. To make
	 * sure that recv() will send us that error code, we tell it not to
	 * use blocking I/O */
	/*while ( recvline_timeo(sock, buffer, sizeof (buffer)-1, 3) > 0);*/
	
	switch (authtype) {
	case (conn_auth_t) PLAIN:
		if (show_dialog) {
			invoke_in_gui_thread (mailthread_update_status_text, "Sending login name...");
		}
		
		/* lets login to the server... */
		if (!pop_user (sock, login, ex)) {
			/* bad login name maybe? I dunno... */
			close (sock);
			return 0;
		}
		
		if (show_dialog) {
			invoke_in_gui_thread (mailthread_update_status_text, "Sending password...");
		}
		
		if (!pop_pass (sock, passwd, ex)) {
			/* bad passwd maybe? I dunno... */
			close (sock);
			return 0;
		}
		
		break;
	case (conn_auth_t) APOP:
		challenge_pos = find_string (buffer, "<");
		if (challenge_pos == -1) {
			/* The server didn't understand APOP!*/
			spruce_exception_setv (ex, SPRUCE_EXCEPTION_SERVICE_CANT_AUTHENTICATE,
					       _("Server doesn't support APOP authentication"));
			close (sock);
			return 0;
		}
		
		if (show_dialog) {
			invoke_in_gui_thread (mailthread_update_status_text, "Logging in with APOP");
		}
		
		if (!pop_apop (sock, login, passwd, buffer + challenge_pos, ex)) {
			close (sock);
			return 0;
		}
		break;
	case (conn_auth_t) CRAM_MD5:
		if (show_dialog) {
			invoke_in_gui_thread (mailthread_update_status_text, "Logging in with CRAM-MD5");
		}
		
		if (!pop_auth (sock, login, passwd, ex)) {
			close (sock);
			return 0;
		}
		break;
	}
	
	if (show_dialog) {
		invoke_in_gui_thread (mailthread_update_status_text, "Sending STAT command...");
	}
	
	/* lets find out how many messages we got on the server */
	if ((num = pop_stat (sock, ex)) < 0) {
		/* something went wrong... */
		close (sock);
		return 0;
	}
	
	uid_init ();
	
	if (num > 0) {
		memset (buffer, 0, sizeof (buffer));
		g_snprintf (buffer, sizeof (buffer)-1, "%d messages on the server, looking for new...", num);
		
		if (show_dialog) {
			invoke_in_gui_thread (mailthread_update_status_text, buffer);
		}
		
		/* remove read UIDs from list */
		if (!(msglist = (gint *)pop_stat_trim (sock, &num, ex))) {
			/* apparently this pop server doesn't support UIDL - damn schmucks */
			msglist = g_malloc0 (sizeof (gint) * num);
			for (i = 1; i <= num; i++)
				*(msglist + i - 1) = i;
		}
	}
	
	memset (buffer, 0, sizeof (buffer));
	g_snprintf (buffer, sizeof (buffer)-1, "%d new messages on the server.", num);
	
	if (show_dialog) {
		invoke_in_gui_thread (mailthread_update_status_text, buffer);
		
		if (num > 0) {
			progress = 0;
			invoke_in_gui_thread ((GtkFunction)mailthread_update_status_progress, (gpointer) &progress);
		} else {
			progress = 1;
			invoke_in_gui_thread ((GtkFunction)mailthread_update_status_progress, (gpointer) &progress);
		}
	}
	
	for (i = 1; i <= num; i++, fetched++) {
		g_snprintf (buffer, sizeof (buffer)-1, "Downloading message %d of %d...", i, num);
		
		d(fprintf (stderr, "%s\n", buffer));
		
		if (show_dialog) {
			invoke_in_gui_thread (mailthread_update_status_text, buffer);
			progress = ((gfloat)i / (gfloat)num);
			invoke_in_gui_thread ((GtkFunction)mailthread_update_status_progress, (gpointer) &progress);
		}
		
		if (!pop_retr (sock, *(msglist + i - 1), finbox, del_yn, ex)) {
			/* eek! an error occured!! */
			g_free (msglist);
			close (sock);
			uid_free ();
			return fetched;
		}
		
		/* delete the message if we are told to do so */
		if (del_yn) {
			if (!pop_dele (sock, *(msglist + i - 1), ex)) {
				/* eek! couldn't clickety-click to empty the recycle bin!!! */
				g_free (msglist);
				close (sock);
				uid_free ();
				return fetched;
			}
		}
	}
	
	/* free allocated memory */
	g_free (msglist);
	
#if 0
	/* this is #if 0'd off because we don't want to call
	 * this cuz it will reset the deleted messages */
	if (!pop_rset (sock, ex)) {
		close (sock);
		return fetched;
	}
#endif
	
	if (!pop_quit (sock, ex)) {
		close (sock);
		uid_free ();
		return fetched;
	}
	
#ifdef WITH_SSL
	closeSSLConnection (sock);
#endif
	
	/* close the socket */
	close (sock);
	
	uid_free ();
	
	return fetched;
}
