/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  Spruce
 *  Copyright (C) 1999-2000 Jeffrey Stedfast
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#ifdef HAVE_CONFIG_H
#	include <config.h>
#endif

#include "sort.h"

#include <string.h>
#include <ctype.h>
#include <time.h>
#include "date.h"
#include "parse.h"

#define d(x) x


gint
gint_compare (gconstpointer a, gconstpointer b)
{
	gint first, second;
	
	g_return_val_if_fail (a, 0);
	g_return_val_if_fail (b, 0);
	
	first = GPOINTER_TO_INT (a);
	second = GPOINTER_TO_INT (b);
	
	if (first < second)
		return -1;
	if (first == second)
		return 0;
	return 1;
}

gint
str_compare (GtkCList *clist, gconstpointer s1, gconstpointer s2)
{
	GtkCListRow *row1, *row2;
	gchar *text1, *text2;
	
	g_return_val_if_fail (clist != NULL, 0);
	g_return_val_if_fail (GTK_IS_CLIST (clist), 0);
	g_return_val_if_fail (s1 != NULL, 0);
	g_return_val_if_fail (s2 != NULL, 0);
	
	row1 = (GtkCListRow *) s1;
	row2 = (GtkCListRow *) s2;
	
	text1 = GTK_CELL_TEXT (row1->cell[clist->sort_column])->text;
	text2 = GTK_CELL_TEXT (row2->cell[clist->sort_column])->text;
	
	return g_strcasecmp (text1, text2);
}

gint
id_compare (GtkCList *clist, gconstpointer s1, gconstpointer s2)
{
	GtkCListRow *row1, *row2;
	gchar *text1, *text2;
	
	g_return_val_if_fail (clist != NULL, 0);
	g_return_val_if_fail (GTK_IS_CLIST (clist), 0);
	g_return_val_if_fail (s1 != NULL, 0);
	g_return_val_if_fail (s2 != NULL, 0);
	
	row1 = (GtkCListRow *) s1;
	row2 = (GtkCListRow *) s2;
	
	text1 = GTK_CELL_TEXT (row1->cell[clist->sort_column])->text;
	text2 = GTK_CELL_TEXT (row2->cell[clist->sort_column])->text;
	
	if (atoi (text1) < atoi (text2))
		return -1;
	if (atoi (text1) > atoi (text2))
		return 1;
	return 0;
}


typedef struct {
	char *name;
	char *address;
} InternetAddress;


static InternetAddress *
internet_address_new_from_string (const gchar *string)
{
	/* We have 3 possibilities...
	   1. "Jeffrey Stedfast" <fejj@helixcode.com>
	   2. fejj@helixcode.com
	   3. <fejj@helixcode.com> (Jeffrey Stedfast)
	*/
	InternetAddress *ia;
	gchar *name = NULL, *address = NULL;
	gchar *ptr, *padding = NULL;
	gboolean in_quotes = FALSE;
	gboolean name_first = FALSE;
	
	g_return_val_if_fail (string != NULL, NULL);
	g_return_val_if_fail (*string != '\0', NULL);
	
	/* look for padding between parts... */
	for (ptr = (gchar *) string; *ptr; ptr++) {
		if (*ptr == '"') {
			in_quotes = !in_quotes;
			name_first = TRUE;
		} else if (!in_quotes && isspace (*ptr)) {
			padding = ptr;
			break;
		}
	}
	
	if (padding) {
		padding--;
		/* we have a type-one or a type-three address */
		if (name_first) {
			name = g_strndup (string, (gint) (padding - string));
			g_strstrip (name);
			
			/* strip off the quotes */
			if (*name) {
				*(name + strlen (name) - 1) = '\0';
				memmove (name, name + 1, strlen (name) - 1);
			} else {
				g_free (name);
				name = NULL;
			}
			
			address = strchr (padding, '<');
			if (address) {
				address++;
				for (ptr = address; *ptr && *ptr != '>'; ptr++);
				
				address = g_strndup (address, (gint) (ptr - address));
			}
		} else {
			address = g_strndup (string, (gint) (padding - string));
			g_strstrip (address);
			
			/* strip off the braces */
			if (*address) {
				*(address + strlen (address) - 1) = '\0';
				memmove (address, address + 1, strlen (address) - 1);
			} else {
				g_free (address);
				address = NULL;
			}
			
			name = strchr (padding, '(');
			if (name) {
				name++;
				in_quotes = FALSE;
				ptr = name;
				
				while (*ptr) {
					if (*ptr == '"')
						in_quotes = !in_quotes;
					else if (!in_quotes && *ptr == ')')
						break;
				}
				
				name = g_strndup (name, (gint) (ptr - name));
			}
		}
	} else {
		/* we have a type-two address */
		address = g_strdup (string);
	}
	
	if (!address) {
		/* the address is the most important part! */
		g_free (name);
		return NULL;
	}
	
	ia = g_new (InternetAddress, 1);
	ia->name = name;
	ia->address = address;
	
	return ia;
}

static void
internet_address_destroy (InternetAddress *ia)
{
	g_return_if_fail (ia != NULL);
	
	g_free (ia->name);
	g_free (ia->address);
	g_free (ia);
}

gint
address_compare (GtkCList *clist, gconstpointer s1, gconstpointer s2)
{
	InternetAddress *ia1, *ia2;
	GtkCListRow *row1, *row2;
	char *address1, *address2;
	gint retval;
	
	g_return_val_if_fail (clist != NULL, 0);
	g_return_val_if_fail (GTK_IS_CLIST (clist), 0);
	g_return_val_if_fail (s1 != NULL, 0);
	g_return_val_if_fail (s2 != NULL, 0);
	
	row1 = (GtkCListRow *) s1;
	row2 = (GtkCListRow *) s2;
	
	address1 = GTK_CELL_TEXT (row1->cell[clist->sort_column])->text;
	address2 = GTK_CELL_TEXT (row2->cell[clist->sort_column])->text;
	
	ia1 = internet_address_new_from_string (address1);
	ia2 = internet_address_new_from_string (address2);
	
	if (!ia1->name && !ia2->name) {
		/* if neither has a name we should compare addresses */
		retval = g_strcasecmp (ia1->address, ia2->address);
	} else {
		if (!ia1->name)
			retval = -1;
		else if (!ia2->name)
			retval = 1;
		else {
			/* FIXME: use Nat's e-western-name parser
			 * so we can compare last name then first */
			retval = g_strcasecmp (ia1->name, ia2->name);
		}
	}
	
	internet_address_destroy (ia1);
	internet_address_destroy (ia2);
	
	return retval;
}

gint
subject_compare (GtkCList *clist, gconstpointer s1, gconstpointer s2)
{
	GtkCListRow *row1, *row2;
	char *subject1, *sub1;
	char *subject2, *sub2;
	
	g_return_val_if_fail (clist != NULL, 0);
	g_return_val_if_fail (GTK_IS_CLIST (clist), 0);
	g_return_val_if_fail (s1 != NULL, 0);
	g_return_val_if_fail (s2 != NULL, 0);
	
	row1 = (GtkCListRow *) s1;
	row2 = (GtkCListRow *) s2;
	
	subject1 = GTK_CELL_TEXT (row1->cell[clist->sort_column])->text;
	subject2 = GTK_CELL_TEXT (row2->cell[clist->sort_column])->text;
	
	/* trim off any "Re:"'s at the beginning of subject1 */
	sub1 = subject1;
	while (!g_strncasecmp (sub1, "Re:", 3)) {
		sub1 += 3;
		/* jump over any spaces */
		for ( ; *sub1 && isspace (*sub1); sub1++);
	}
	
	/* trim off any "Re:"'s at the beginning of subject2 */
	sub2 = subject2;
	while (!g_strncasecmp (sub2, "Re:", 3)) {
		sub2 += 3;
		/* jump over any spaces */
		for ( ; *sub2 && isspace (*sub2); sub2++);
	}
	
	/* jump over any spaces */
	for ( ; *sub1 && isspace (*sub1); sub1++);
	for ( ; *sub2 && isspace (*sub2); sub2++);
	
	return g_strcasecmp (sub1, sub2);
}

gint
date_compare (GtkCList *clist, gconstpointer s1, gconstpointer s2)
{
	/* Original function by Jussi Junni, modified by Jeffrey Stedfast */
	GtkCListRow *row1, *row2;
	gchar *text1, *text2;
	date_t date1, date2;
	
	g_return_val_if_fail (clist != NULL, 0);
	g_return_val_if_fail (GTK_IS_CLIST (clist), 0);
	g_return_val_if_fail (s1 != NULL, 0);
	g_return_val_if_fail (s2 != NULL, 0);
	
	row1 = (GtkCListRow *) s1;
	row2 = (GtkCListRow *) s2;
	
	text1 = GTK_CELL_TEXT (row1->cell[clist->sort_column])->text;
	text2 = GTK_CELL_TEXT (row2->cell[clist->sort_column])->text;
	
	if (text2 == NULL)
		return (text1 != NULL);
	if (text1 == NULL)
		return -1;
	
	date1 = parse_date (text1);
	date2 = parse_date (text2);
	
	if (date1.year < date2.year) return 1;
	if (date1.year > date2.year) return -1;
	
	if (date1.month != 0 && date2.month != 0) {
		if (date1.month < date2.month) return 1;
		if (date1.month > date2.month) return -1;
	} else {
		d(fprintf (stderr, "Invalid date format! (%s)\n", date1.month == 0 ? text1 : text2));
	}
	
	if (date1.day < date2.day) return 1;
	if (date1.day > date2.day) return -1;
	
	if (date1.hour < date2.hour) return 1;
	if (date1.hour > date2.hour) return -1;
	
	if (date1.minute < date2.minute) return 1;
	if (date1.minute > date2.minute) return -1;
	
	if (date1.second < date2.second) return 1;
	if (date1.second > date2.second) return -1;
	
	return 0;
}
