/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*  spruce-ssl.c - SSL support for Spruce
 *  Copyright (C) 2000 Erik Forsberg
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include "spruce-ssl.h"

#include <sys/types.h>
#include <sys/socket.h>

#ifdef WITH_SSL
#include "server.h"

#define d(x) x

static struct sslPair *sslPairs;

/* Put SSL structure in sorted linked list, together with socket number */
void registerSSL (int sock, SSL *sslStruct) 
{
	if (!sslPairs) {
		sslPairs = g_malloc (sizeof (struct sslPair));
		sslPairs->next = NULL;
		sslPairs->sock = sock;
		sslPairs->sslStruct = sslStruct;
		
		return;
	} else {
		struct sslPair *cur = sslPairs;
		
		while (NULL != cur->next && cur->next->sock < sock) {
			cur=cur->next;
		}
		
		if (!cur->next) {
			cur->next = g_malloc (sizeof (struct sslPair));
			cur->next->next = NULL;
			cur = cur->next;
		} else {
			struct sslPair *oldNext = cur->next;
			
			cur = g_malloc (sizeof (struct sslPair));
			cur->next = oldNext;
		}
		
		cur->sock = sock;
		cur->sslStruct = sslStruct;
	}
}

/* Get SSL structure for a socket number */
SSL *getSSL(int sock) 
{
	struct sslPair *cur = sslPairs;
	
	while (cur && cur->sock < sock) {
		cur = cur->next;
	}
	
	if (cur) 
		return cur->sslStruct;
	
	return NULL;
}

/* Delete SSL structure from linked list */  
void deregisterSSL(int sock) 
{
	struct sslPair *cur, *old;
	
	cur = sslPairs;
	old = NULL;
	
	while (cur && cur->sock != sock) {
		if (cur != sslPairs) 
			old = cur;
		cur=cur->next;
	}
	
	if (cur) {
		if (old)
			old->next = cur->next;
		if (cur->sslStruct)
			g_free (cur->sslStruct);
		g_free (cur);
	}
}

/* Initialize some ssl structures */
void sslInit (void) 
{
	SSL_load_error_strings ();
	SSLeay_add_ssl_algorithms ();
}

int verifyCallback (int ok, X509_STORE_CTX *ctx) 
{
	X509 *cert;
	int err;
	
	cert = X509_STORE_CTX_get_current_cert (ctx);
	err = X509_STORE_CTX_get_error(ctx);
	
	d(fprintf (stderr, "Error verifying %s\n", 
		   X509_NAME_oneline (X509_get_subject_name (cert), 0, 0)));
	d(fprintf (stderr, "Error: number %d: %s\n", 
		   err, X509_verify_cert_error_string (err)));
	
	return 1;
}


/* Open a SSL connection. Called after connect */
int openSSLConnection (gint sockfd) 
{
	int n;
	SSL_CTX *ssl_ctx = NULL;
	SSL *ssl = NULL;
	
	ssl_ctx = SSL_CTX_new (SSLv23_client_method ());
	SSL_CTX_set_verify (ssl_ctx, SSL_VERIFY_PEER, &verifyCallback);
	ssl = SSL_new (ssl_ctx);
	SSL_set_fd (ssl, sockfd);
	
	n = SSL_connect (ssl);
	if (!(n < 0)) {
		registerSSL (sockfd, ssl);
	}
	
	return (n);
}

/* Shut down SSL connection. Called before shutdown */
void closeSSLConnection (gint sockfd) 
{
	SSL *ssl = getSSL (sockfd);
	
	if (!ssl) /* Not a SSL connection */
		return; 
	
	deregisterSSL (sockfd);
	if (!ssl)
		return;
	
	SSL_shutdown (ssl);
	
	if (ssl && ssl->ctx)
		SSL_CTX_free (ssl->ctx);
	
	if (ssl) free (ssl);
}

int verifySSLConnection (gint sockfd) 
{
	int res;
	STACK_OF (X509) *chain;
	SSL *ssl = getSSL (sockfd);
	
	chain = SSL_get_peer_cert_chain (ssl);
	res = ssl_verify_cert_chain (ssl, chain);
	
	return (res);
}


#ifdef TEST_SSL

int main (void) 
{
	gint sock;
	int n;
	char buf[1024];
	char *p;
	SSL *ssl;
	Server server;
	
	server.hostname = g_strdup ("imap.lysator.liu.se");
	server.port = 995;
	sslInit ();
	
	if (!Resolve (&server)) {
		d(fprintf (stderr, "Couldn't resolve hostname\n"));
		return -1;
	}
	
	server.sin.sin_family = AF_INET;
	server.sin.sin_port = htons (server.port);
	sock = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
	
	if (sock < 0) {
		perror("socket ()");
		return -1;
	}
	
	n = connect (sock, &server.sin, sizeof (server.sin));
	if (n < 0) {
		perror("connect ()");
		return -1;
	}
	
	n = openSSLConnection (sock);
	/*  if (!verifySSLConnection (sock)) {
	    fprintf (stderr, "Could not verify SSL certificate\n");
	    }*/
	if (n < 0) {
		d(fprintf (stderr, "openSSLConnection returned %d\n", n));
		return -1;
	}
	ssl = getSSL (sock);
	if (!ssl) {
		d(fprintf (stderr, "Hmm.. couldn't get ssl structure for socket %d\n", sock));
		return -1;
	}
	
	p = (unsigned char *)buf;
	n = SSL_read (ssl, p, 1024);
	
	d(fprintf (stderr, "Read %d bytes\n", n));
	d(fprintf (stderr, "%s\n", buf));
	
	closeSSLConnection (sock);
	shutdown (sock, 2);
	close (sock);
	
	return 0;
}


#endif  /* TEST_SSL */
#endif /* WITH_SSL */

ssize_t SSLSend (gint socket, gchar *buffer, gint buflen, gint flags) 
{
#ifdef WITH_SSL
	SSL *ssl = getSSL(socket);
	if (ssl) 
		return (SSL_write (ssl, buffer, buflen));
	else
#endif
		return (send (socket, buffer, buflen, flags));
}


