#!/usr/local/bin/perl
#
#
# import_addrbook.pl
#
# This script imports pine's address book into spruce mailboxes.
# 
# If you are a pine user and wish to start using spruce as your mail client
# or simply try out spruce and test how it looks, this is for you. You
# can convert existing pine address book into spruce address book. 
#
# Command line Usage: 
#   import_addrbook.pl in_pine_addressbook > out_spruce_addressbook
# 
# Ex.: 
#   import_addrbook.pl /home/roychri/.adressbook > ~/.spruce/addrbook
#
# This will use your pine address book file and create one new file
# out of it. 
# 
# Installation
#
# 1) Put this file in your path
# 2) Make it executable.
# 3) Check to make sure it points to the right perl interpreter
# 4) Run it! (See the command line usage info above)
#
# Documentation
#
# Look in the source code to see how it really works.
# There's also POD documentation available within this script.
#
#  To see that POD documentation run
#    pod2text import_addrbook.pl | more
#
#  To generate an external file for the POD documentation
#    pod2text import_addrbook.pl > import_addrbook.doc
#
#  To generate a manpage from that POD
#    pod2man import_addrbook.pl > import_addrbook.1
#
# Author: Christian Roy <roychri@infiniweb.ca>
#         President
#         Station InfiniWeb Inc - http://www.infiniweb.ca
# 

=head1 NAME

 import_addrbook.pl - Pine address book importer for spruce

=head1 SYNOPSIS

 import_addrbook.pl <in_adressbook> > <out_adressbook>

=head1 VERSION

Current version is: 0.1

Release status is: alpha

Release date is: June 8th, 2000

=head1 DESCRIPTION

 This script imports pine's address book into spruce mailboxes.
 
 If you are a pine user and wish to start using spruce as your mail client
 or simply try out spruce and test how it looks, this is for you. You
 can convert existing pine address book into spruce address book. 

=head1 INSTALLATION

Follow these simple steps :

=over 4

=item 1) Put this file in your path

 ex.: mv import_addrbook.pl /usr/local/bin

=item 2) Make it executable.

 ex.: chmod 755 import_addrbook.pl

=item 3) Check shebang line

 Check to make sure it points to the right perl interpreter

=item 4) Run it!

 (See the synopsis above)

=back


=head1 REQUIREMENTS

This script requires the following perl modules to be installed :

=over 4

=item Mail::Address

 Mailbox access via the c-client library API

=back

Those modules can be found at http://www.cpan.org

=head1 DOCUMENTATION

 The documentation and upgdare information is also available at :

 http://projects.infiniweb.ca/spruce

 or look in the source code to see how it really works.

=head1 TO DO

=over 4

=item Handle existing entries

 Do not overwrite existing spruce adressbook, simply add to it. (Or prompt for action)

=back

=head1 KNOWN BUGS

 none. (yet) :)

=head1 SEE ALSO

 L<perl>, L<import_mbox>

=head1 COPYRIGHT

Copyright 2000 Station InfiniWeb Inc.

This script is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 AUTHOR

Christian Roy <roychri@infiniweb.ca>

=cut


use strict;
use Mail::Address;

my $tempon = undef;
my $rowid = 1;
my @AddrBook;

while(<>) {
    chomp;
    next if(/^$/);
    my $entry;
    if(/^(\S+)\t([^\t]+)\t([^\t]+)$/) {
	my @addrs = Mail::Address->parse($3);
	warn("cannot parse address"),next if($#addrs != 0);
	my $name = $2;
	$name = $addrs[0]->phrase if(!$name || $name eq "");
	$entry = new AddrBookEntry('handle'=>$1,'name'=>$name,
				   'email'=>$addrs[0]->address);
	$tempon = undef;
    } elsif(/\s<?([^\@]+\@\S+)>?$/) {
	$tempon .= $_;
	if($tempon =~ /^(\S+)\t([^\t]*)\t([^\t]+)$/) {
	    my @addrs = Mail::Address->parse($3);
	    warn("cannot parse address"),next if($#addrs != 0);
	    my $name = $2;
	    $name = $addrs[0]->phrase if(!$name || $name eq "");
	    $entry = new AddrBookEntry('handle'=>$1,'name'=>$name,
				       'email'=>$addrs[0]->address);
	} else {
	    warn("Invalid entry");
	}
	$tempon = undef;
    } else {
	$tempon .= $_;
    }
    if($entry && !$tempon) {
	push(@AddrBook,$entry);
	$entry = undef;
    } elsif($entry) {
	warn("Warning: Internal error... An entry was created while buffer still active.");
	$entry = undef;
    }
    $rowid++;
}

my $newbook = new Spruce_AddrBook('book'=>\@AddrBook);
$newbook->writeOut();

package Spruce_AddrBook;

sub new {
    my $class = shift;
    my %param = @_;
    my $self = {
	'book'=>$param{'book'},
	};
    bless $self,$class;
    return $self;
}

sub writeOut {
    my $self = shift;
    return if(!($self->{'book'}));
    my @book = @{$self->{'book'}};
    my $rowid = 0;
    print "# This is a generated file made by Spruce.\n";
    print "# Please do NOT edit this by hand, please instead use\n";
    print "# Spruce's address book editor to modify this file.\n\n";
    foreach my $entry (@book) {
	print "# Address $rowid\n";
	print "Name: ".$entry->{'name'}."\n";
	print "Email: ".$entry->{'email'}."\n";
	print "Memo: \n\n";
	$rowid++;
    }
}


package AddrBookEntry;

sub new {
    my $class = shift;
    my %param = @_;
    my $self = {
	'handle'=>$param{'handle'},
	'name'=>$param{'name'},
	'email'=>$param{'email'},
    };
    bless $self,$class;
    return $self;
}
