#!/usr/local/bin/perl -w
#
#
# import_mbox.pl
#
# This script imports c-client compatible mailboxes (The c-client
# library is the core library used by Pine and the University of
# Washington IMAP daemon [written by Mark Crispin]) into spruce mailboxes.
# 
# If you are a pine user and wish to start using spruce as your mail client
# or simply try out spruce and test how it looks, this is for you. You
# can convert existing pine mailboxes into spruce mailboxes. If you
# only have one huge pine mailbox, then use this program only once.
# If you have many different pine mailboxes, you may use this program
# for each mailboxes that you wish to convert, specifying different output
# spruce files for each mailboxes.
#
# Command line Usage: 
#   import_mbox.pl in_pine_mailbox_path out_spruce_mailbox out_spruce_index
# 
# Ex.: 
#   import_mbox.pl /home/roychri/mbox ~/.spruce/Inbox.mbx ~/.spruce/Inbox.idx
#
# This will use your pine Inbox mailbox file and create two new files 
# out of it. If you want to convert other mailboxes, you will have to create
# them as Folders using the spruce interface before you can start using it.
# 
# Installation
#
# 1) Put this file in your path
# 2) Make it executable.
# 3) Check to make sure it points to the right perl interpreter
# 4) Run it! (See the command line usage info above)
#
# Documentation
#
# Look in the source code to see how it really works.
# There's also POD documentation available within this script.
#
#  To see that POD documentation run
#    pod2text import_mbox.pl | more
#
#  To generate an external file for the POD documentation
#    pod2text import_mbox.pl > import_mbox.doc
#
#  To generate a manpage from that POD
#    pod2man import_mbox.pl > import_mbox.1
#
# Author: Christian Roy <roychri@infiniweb.ca>
#         President
#         Station InfiniWeb Inc - http://www.infiniweb.ca
# 

=head1 NAME

 import_mbox.pl - Pine mailbox importer

=head1 SYNOPSIS

 import_mbox.pl <in_mbbox> <out_mbox> <out_index>

=head1 VERSION

Current version is: 0.1

Release status is: alpha

Release date is: June 8th, 2000

=head1 DESCRIPTION

This script imports c-client compatible mailboxes (The c-client
library is the core library used by Pine and the University of
Washington IMAP daemon [written by Mark Crispin]) into spruce mailboxes.
 
If you are a pine user and wish to start using spruce as your mail client
or simply try out spruce and test how it looks, this is for you. You
can convert existing pine mailboxes into spruce mailboxes. If you
only have one huge pine mailbox, then use this program only once.
If you have many different pine mailboxes, you may use this program
for each mailboxes that you wish to convert, specifying diffrent output
spruce files for each mailboxes.

=head1 INSTALLATION

Follow these simple steps :

=over 4

=item 1) Put this file in your path

 ex.: mv import_mbox.pl /usr/local/bin

=item 2) Make it executable.

 ex.: chmod 755 import_mbox.pl

=item 3) Check shebang line

 Check to make sure it points to the right perl interpreter

=item 4) Run it!

 (See the synopsis above)

=back


=head1 REQUIREMENTS

This script requires the following perl modules to be installed :

=over 4

=item Mail::Cclient

 Mailbox access via the c-client library API

=item MIME::QuotedPrint

 Encoding and decoding of quoted-printable strings

=item Mail::Header

 manipulate mail RFC822 compliant headers

=back

Those modules can be found at http://www.cpan.org

=head1 DOCUMENTATION

 The documentation and upgdare information is also available at :

 http://projects.infiniweb.ca/spruce

 or look in the source code to see how it really works.

=head1 TO DO

=over 4

=item Guess mailbox and index filename just with supplied name.

Instead of supplying the path name, supply name of mailbox and will guess the appropriate filenames. This could be an optional way of using the script, to keep allowing specifying the fill path names of both mailbox and index file.

=item spruce Configuration check

Go see in spruce's configuration to see if it's a valid mailbox.

=item Optional spruce config change.

Automaticly add a new spruce mailbox (Folder) when specifying one that doesn't exists.

=back

=head1 KNOWN BUGS

 none. (yet) :)

=head1 SEE ALSO

 L<perl>

=head1 COPYRIGHT

Copyright 2000 Station InfiniWeb Inc.

This script is free software; you can redistribute it and/or
modify it under the same terms as Perl itself.

=head1 AUTHOR

Christian Roy <roychri@infiniweb.ca>

=cut


use strict;
use Mail::Cclient;
use MIME::QuotedPrint;
use Mail::Header;

die("Invalid number of parameters") if($#ARGV != 2);

my $mbox_input = shift(@ARGV);
my $mbox_file = shift(@ARGV);
my $index_file = shift(@ARGV);

open(MBX,">$mbox_file") || die("unable to open $mbox_file $!");
open(IDX,">$index_file") || die("unable to open $index_file $!");

# Read your mailbox file.
my $c = Mail::Cclient->new($mbox_input, 'readonly');

if($c) {
    my $num_msgs = $c->nmsgs();	# Get number of messages.
    my $offset = 0;		# Keep track of offset.
    foreach my $msgid (1..$num_msgs) { # For each messages...
	my($subject,$sender,$seen,$date,$size,$hdr,$text);
	my ($envelope, $body) = $c->fetchstructure($msgid);
	my $elt = $c->elt($msgid);

	# Handles header to get Subject, Sender and Date.
	my $raw_header = $c->fetchheader($msgid);
	my @arr_header = split(/\r?\n/,$raw_header);
	my $header = new Mail::Header(\@arr_header);
	# Spruce's mailboxes worked with Quoted Printable text.
	$header->replace('Content-type','TEXT/PLAIN; charset=iso-8859-1');
	$header->replace('Content-transfer-encoding','QUOTED-PRINTABLE');
	$hdr = $header->as_string."\n";

	# Encode the text 'Quote-Printable'
	$text = encode_qp($c->fetchtext($msgid));

	$seen = 0;		# We should use the 'Status:' header instead.
	$size = length($hdr) + length($text) + 2; # Real length (once encoded)
	$subject = $envelope->subject;
	# Let's handle only one sender (Even though it's possible to have more)
	my @senders = @{$envelope->sender};
	if(@senders && $senders[0]->personal) {
	    $sender = $senders[0]->personal." <".$senders[0]->mailbox."\@".$senders[0]->host.">";
	} elsif(@senders) {
	    $sender = $senders[0]->mailbox."\@".$senders[0]->host;
	}
	$date = $elt->date;

	# Output stuff into files
	select MBX;
	print "$hdr$text.\r\n";	# Mailbox file
	select IDX;
	print "$offset\n";	# Index file
	print "$size\n";
	print "$seen\n";
	print "$sender\n";
	print "$subject\n";
	print "$date\n";
	select STDOUT;
	$offset += $size;	# Keep track of offset
    }
    print "Number of messages : $num_msgs\n";
} else {
    print "Unable to open mailbox $mbox_input\n";
}
